# -----------------------------------------------------------------------------
# Copyright (c) 2023, 2024, Oracle and/or its affiliates.
#
# This software is dual-licensed to you under the Universal Permissive License
# (UPL) 1.0 as shown at https://oss.oracle.com/licenses/upl and Apache License
# 2.0 as shown at http://www.apache.org/licenses/LICENSE-2.0. You may choose
# either license.
#
# If you elect to accept the software under the Apache License, Version 2.0,
# the following applies:
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# -----------------------------------------------------------------------------

"""
6300 - Module for testing other cursor methods and attributes with asyncio.
"""

import unittest

import oracledb
import test_env


@unittest.skipUnless(
    test_env.get_is_thin(), "asyncio not supported in thick mode"
)
class TestCase(test_env.BaseAsyncTestCase):
    async def test_6300(self):
        "6300 - test preparing a statement and executing it multiple times"
        cursor = self.conn.cursor()
        self.assertEqual(cursor.statement, None)
        statement = "begin :value := :value + 5; end;"
        cursor.prepare(statement)
        var = cursor.var(oracledb.NUMBER)
        self.assertEqual(cursor.statement, statement)
        var.setvalue(0, 2)
        await cursor.execute(None, value=var)
        self.assertEqual(var.getvalue(), 7)
        await cursor.execute(None, value=var)
        self.assertEqual(var.getvalue(), 12)
        await cursor.execute("begin :value2 := 3; end;", value2=var)
        self.assertEqual(var.getvalue(), 3)

    async def test_6301(self):
        "6301 - confirm an exception is raised after closing a cursor"
        self.cursor.close()
        with self.assertRaisesFullCode("DPY-1006"):
            await self.cursor.execute("select 1 from dual")

    async def test_6302(self):
        "6302 - test iterators"
        await self.cursor.execute(
            """
            select IntCol
            from TestNumbers
            where IntCol between 1 and 3
            order by IntCol
            """
        )
        rows = [v async for v, in self.cursor]
        self.assertEqual(rows, [1, 2, 3])

    async def test_6303(self):
        "6303 - test iterators (with intermediate execute)"
        await self.cursor.execute("truncate table TestTempTable")
        await self.cursor.execute(
            """
            select IntCol
            from TestNumbers
            where IntCol between 1 and 3
            order by IntCol
            """
        )
        test_iter = self.cursor.__aiter__()
        (value,) = await test_iter.__anext__()
        await self.cursor.execute(
            "insert into TestTempTable (IntCol) values (1)"
        )
        with self.assertRaisesFullCode("DPY-1003"):
            await test_iter.__anext__()

    async def test_6304(self):
        "6304 - test setting input sizes without any parameters"
        self.cursor.setinputsizes()
        await self.cursor.execute("select :val from dual", val="Test Value")
        self.assertEqual(await self.cursor.fetchall(), [("Test Value",)])

    async def test_6305(self):
        "6305 - test setting input sizes with an empty dictionary"
        empty_dict = {}
        self.cursor.prepare("select 236 from dual")
        self.cursor.setinputsizes(**empty_dict)
        await self.cursor.execute(None, empty_dict)
        self.assertEqual(await self.cursor.fetchall(), [(236,)])

    async def test_6306(self):
        "6306 - test setting input sizes with an empty list"
        empty_list = []
        self.cursor.prepare("select 239 from dual")
        self.cursor.setinputsizes(*empty_list)
        await self.cursor.execute(None, empty_list)
        self.assertEqual(await self.cursor.fetchall(), [(239,)])

    async def test_6307(self):
        "6307 - test setting input sizes with positional args"
        var = self.cursor.var(oracledb.STRING, 100)
        self.cursor.setinputsizes(None, 5, None, 10, None, oracledb.NUMBER)
        await self.cursor.execute(
            """
            begin
              :1 := :2 || to_char(:3) || :4 || to_char(:5) || to_char(:6);
            end;
            """,
            [var, "test_", 5, "_second_", 3, 7],
        )
        self.assertEqual(var.getvalue(), "test_5_second_37")

    async def test_6308(self):
        "6308 - test parsing query statements"
        sql = "select LongIntCol from TestNumbers where IntCol = :val"
        await self.cursor.parse(sql)
        self.assertEqual(self.cursor.statement, sql)
        self.assertEqual(
            self.cursor.description,
            [("LONGINTCOL", oracledb.DB_TYPE_NUMBER, 17, None, 16, 0, 0)],
        )

    async def test_6309(self):
        "6309 - test binding boolean data without the use of PL/SQL"
        await self.cursor.execute("truncate table TestTempTable")
        sql = "insert into TestTempTable (IntCol, StringCol1) values (:1, :2)"
        await self.cursor.execute(sql, (False, "Value should be 0"))
        await self.cursor.execute(sql, (True, "Value should be 1"))
        await self.cursor.execute(
            "select IntCol, StringCol1 from TestTempTable order by IntCol"
        )
        expected_value = [(0, "Value should be 0"), (1, "Value should be 1")]
        self.assertEqual(await self.cursor.fetchall(), expected_value)

    async def test_6310(self):
        "6310 - test using a cursor as a context manager"
        with self.cursor as cursor:
            await cursor.execute("truncate table TestTempTable")
            await cursor.execute("select count(*) from TestTempTable")
            (count,) = await cursor.fetchone()
            self.assertEqual(count, 0)
        with self.assertRaisesFullCode("DPY-1006"):
            self.cursor.close()

    async def test_6311(self):
        "6311 - test that rowcount attribute is reset to zero on query execute"
        for num in [0, 1, 1, 0]:
            await self.cursor.execute("select * from dual where 1 = :s", [num])
            await self.cursor.fetchone()
            self.assertEqual(self.cursor.rowcount, num)

    async def test_6312(self):
        "6312 - test that an object type can be used as type in cursor.var()"
        obj_type = await self.conn.gettype("UDT_OBJECT")
        var = self.cursor.var(obj_type)
        await self.cursor.callproc(
            "pkg_TestBindObject.BindObjectOut", (28, "Bind obj out", var)
        )
        obj = var.getvalue()
        result = await self.cursor.callfunc(
            "pkg_TestBindObject.GetStringRep", str, (obj,)
        )
        exp = "udt_Object(28, 'Bind obj out', null, null, null, null, null)"
        self.assertEqual(result, exp)

    async def test_6313(self):
        "6313 - test that fetching an XMLType returns a string"
        int_val = 5
        label = "IntCol"
        expected_result = f"<{label}>{int_val}</{label}>"
        await self.cursor.execute(
            f"""
            select XMLElement("{label}", IntCol)
            from TestStrings
            where IntCol = :int_val
            """,
            int_val=int_val,
        )
        (result,) = await self.cursor.fetchone()
        self.assertEqual(result, expected_result)

    async def test_6314(self):
        "6314 - test last rowid"

        # no statement executed: no rowid
        self.assertIsNone(self.cursor.lastrowid)

        # DDL statement executed: no rowid
        await self.cursor.execute("truncate table TestTempTable")
        self.assertIsNone(self.cursor.lastrowid)

        # statement prepared: no rowid
        self.cursor.prepare("insert into TestTempTable (IntCol) values (:1)")
        self.assertIsNone(self.cursor.lastrowid)

        # multiple rows inserted: rowid of last row inserted
        rows = [(n,) for n in range(225)]
        await self.cursor.executemany(None, rows)
        rowid = self.cursor.lastrowid
        await self.cursor.execute(
            """
            select rowid
            from TestTempTable
            where IntCol = :1
            """,
            rows[-1],
        )
        row = await self.cursor.fetchone()
        self.assertEqual(row[0], rowid)

        # statement executed but no rows updated: no rowid
        await self.cursor.execute("delete from TestTempTable where 1 = 0")
        self.assertIsNone(self.cursor.lastrowid)

        # stetement executed with one row updated: rowid of updated row
        await self.cursor.execute(
            """
            update TestTempTable set StringCol1 = 'Modified'
            where IntCol = :1
            """,
            rows[-2],
        )
        rowid = self.cursor.lastrowid
        await self.cursor.execute(
            "select rowid from TestTempTable where IntCol = :1",
            rows[-2],
        )
        row = await self.cursor.fetchone()
        self.assertEqual(row[0], rowid)

        # statement executed with many rows updated: rowid of last updated row
        await self.cursor.execute(
            """
            update TestTempTable set
                StringCol1 = 'Row ' || to_char(IntCol)
            where IntCol = :1
            """,
            rows[-3],
        )
        rowid = self.cursor.lastrowid
        await self.cursor.execute(
            "select StringCol1 from TestTempTable where rowid = :1",
            [rowid],
        )
        row = await self.cursor.fetchone()
        self.assertEqual(row[0], "Row %s" % rows[-3])

    async def test_6315(self):
        "6315 - test prefetch rows"
        await self.setup_round_trip_checker()

        # perform simple query and verify only one round trip is needed
        with self.conn.cursor() as cursor:
            await cursor.execute("select sysdate from dual")
            await cursor.fetchall()
            await self.assertRoundTrips(1)

        # set prefetchrows to 1 and verify that two round trips are now needed
        with self.conn.cursor() as cursor:
            cursor.prefetchrows = 1
            self.assertEqual(cursor.prefetchrows, 1)
            await cursor.execute("select sysdate from dual")
            await cursor.fetchall()
            await self.assertRoundTrips(2)

        # simple DDL only requires a single round trip
        with self.conn.cursor() as cursor:
            await cursor.execute("truncate table TestTempTable")
            await self.assertRoundTrips(1)

        # array execution only requires a single round trip
        num_rows = 590
        with self.conn.cursor() as cursor:
            data = [(n + 1,) for n in range(num_rows)]
            await cursor.executemany(
                "insert into TestTempTable (IntCol) values (:1)",
                data,
            )
            await self.assertRoundTrips(1)

        # setting prefetch and array size to 1 requires a round-trip for each
        # row
        with self.conn.cursor() as cursor:
            cursor.arraysize = 1
            cursor.prefetchrows = 1
            self.assertEqual(cursor.prefetchrows, 1)
            await cursor.execute("select IntCol from TestTempTable")
            await cursor.fetchall()
            await self.assertRoundTrips(num_rows + 1)

        # setting prefetch and array size to 300 requires 2 round-trips
        with self.conn.cursor() as cursor:
            cursor.arraysize = 300
            cursor.prefetchrows = 300
            self.assertEqual(cursor.prefetchrows, 300)
            await cursor.execute("select IntCol from TestTempTable")
            await cursor.fetchall()
            await self.assertRoundTrips(2)

    async def test_6316(self):
        "6316 - test prefetch rows using existing cursor"
        await self.setup_round_trip_checker()

        # Set prefetch rows on an existing cursor
        num_rows = 590
        with self.conn.cursor() as cursor:
            await cursor.execute("truncate table TestTempTable")
            await self.assertRoundTrips(1)
            data = [(n + 1,) for n in range(num_rows)]
            await cursor.executemany(
                "insert into TestTempTable (IntCol) values (:1)",
                data,
            )
            await self.assertRoundTrips(1)
            cursor.prefetchrows = 30
            cursor.arraysize = 100
            await cursor.execute("select IntCol from TestTempTable")
            await cursor.fetchall()
            await self.assertRoundTrips(7)

    async def test_6317(self):
        "6317 - test parsing plsql statements"
        sql = "begin :value := 5; end;"
        await self.cursor.parse(sql)
        self.assertEqual(self.cursor.statement, sql)
        self.assertIsNone(self.cursor.description)

    async def test_6318(self):
        "6318 - test parsing ddl statements"
        sql = "truncate table TestTempTable"
        await self.cursor.parse(sql)
        self.assertEqual(self.cursor.statement, sql)
        self.assertIsNone(self.cursor.description)

    async def test_6319(self):
        "6319 - test parsing dml statements"
        sql = "insert into TestTempTable (IntCol) values (1)"
        await self.cursor.parse(sql)
        self.assertEqual(self.cursor.statement, sql)
        self.assertIsNone(self.cursor.description)

    async def test_6320(self):
        "6320 - test binding by name with leading colon"
        params = {":arg1": 5}
        await self.cursor.execute("select :arg1 from dual", params)
        (result,) = await self.cursor.fetchone()
        self.assertEqual(result, params[":arg1"])

    async def test_6321(self):
        "6321 - test binding mixed null and not null values in a PL/SQL block"
        out_vars = [self.cursor.var(str) for i in range(4)]
        await self.cursor.execute(
            """
            begin
                :1 := null;
                :2 := 'Value 1';
                :3 := null;
                :4 := 'Value 2';
            end;
            """,
            out_vars,
        )
        values = [var.getvalue() for var in out_vars]
        self.assertEqual(values, [None, "Value 1", None, "Value 2"])

    async def test_6322(self):
        "6322 - test excluding statement from statement cache"
        num_iters = 10
        sql = "select user from dual"
        await self.setup_parse_count_checker()

        # with statement cache enabled, only one parse should take place
        for i in range(num_iters):
            with self.conn.cursor() as cursor:
                await cursor.execute(sql)
        await self.assertParseCount(1)

        # with statement cache disabled for the statement, parse count should
        # be the same as the number of iterations
        for i in range(num_iters):
            with self.conn.cursor() as cursor:
                cursor.prepare(sql, cache_statement=False)
                await cursor.execute(None)
        await self.assertParseCount(num_iters - 1)

    async def test_6323(self):
        "6323 - test repeated DDL"
        await self.cursor.execute("truncate table TestTempTable")
        await self.cursor.execute(
            "insert into TestTempTable (IntCol) values (1)"
        )
        await self.cursor.execute("truncate table TestTempTable")
        await self.cursor.execute(
            "insert into TestTempTable (IntCol) values (1)"
        )

    async def test_6324(self):
        "6324 - test executing SQL with non-ASCII characters"
        await self.cursor.execute("select 'FÖÖ' from dual")
        (result,) = await self.cursor.fetchone()
        self.assertIn(result, ("FÖÖ", "F¿¿"))

    async def test_6325(self):
        "6325 - test case sensitivity of unquoted bind names"
        await self.cursor.execute("select :test from dual", {"TEST": "a"})
        (result,) = await self.cursor.fetchone()
        self.assertEqual(result, "a")

    async def test_6326(self):
        "6326 - test case sensitivity of quoted bind names"
        with self.assertRaisesFullCode("ORA-01036", "DPY-4008"):
            await self.cursor.execute(
                'select :"test" from dual', {'"TEST"': "a"}
            )

    async def test_6327(self):
        "6327 - test using a reserved keywords as a bind name"
        sql = "select :ROWID from dual"
        with self.assertRaisesFullCode("ORA-01745"):
            await self.cursor.parse(sql)

    async def test_6328(self):
        "6328 - test array size less than prefetch rows"
        for i in range(2):
            with self.conn.cursor() as cursor:
                cursor.arraysize = 1
                await cursor.execute(
                    "select 1 from dual union select 2 from dual"
                )
                self.assertEqual(await cursor.fetchall(), [(1,), (2,)])

    async def test_6329(self):
        "6329 - test re-executing a query with blob as bytes"

        def type_handler(cursor, metadata):
            if metadata.type_code is oracledb.DB_TYPE_BLOB:
                return cursor.var(bytes, arraysize=cursor.arraysize)

        self.conn.outputtypehandler = type_handler
        blob_data = b"An arbitrary set of blob data for test case 4348"
        await self.cursor.execute("truncate table TestBLOBs")
        await self.cursor.execute(
            "insert into TestBLOBs (IntCol, BlobCol) values (1, :data)",
            [blob_data],
        )
        await self.cursor.execute("select IntCol, BlobCol from TestBLOBs")
        self.assertEqual(await self.cursor.fetchall(), [(1, blob_data)])

        await self.cursor.execute("truncate table TestBLOBs")
        await self.cursor.execute(
            "insert into TestBLOBs (IntCol, BlobCol) values (1, :data)",
            [blob_data],
        )
        await self.cursor.execute("select IntCol, BlobCol from TestBLOBs")
        self.assertEqual(await self.cursor.fetchall(), [(1, blob_data)])

    async def test_6330(self):
        "6330 - test re-executing a statement after raising an error"
        sql = "select * from TestFakeTable"
        with self.assertRaisesFullCode("ORA-00942"):
            await self.cursor.execute(sql)
        with self.assertRaisesFullCode("ORA-00942"):
            await self.cursor.execute(sql)

        sql = "insert into TestStrings (StringCol) values (NULL)"
        with self.assertRaisesFullCode("ORA-01400"):
            await self.cursor.execute(sql)
        with self.assertRaisesFullCode("ORA-01400"):
            await self.cursor.execute(sql)

    async def test_6331(self):
        "6331 - test executing a statement that raises ORA-01007"
        with self.conn.cursor() as cursor:
            await cursor.execute(
                """
                create or replace view ora_1007 as
                    select 1 as SampleNumber, 'String' as SampleString,
                        'Another String' as AnotherString
                    from dual
                """
            )
        with self.conn.cursor() as cursor:
            await cursor.execute("select * from ora_1007")
            self.assertEqual(
                await cursor.fetchone(), (1, "String", "Another String")
            )
        with self.conn.cursor() as cursor:
            await cursor.execute(
                """
                create or replace view ora_1007 as
                    select 1 as SampleNumber,
                        'Another String' as AnotherString
                    from dual
                """
            )
        with self.conn.cursor() as cursor:
            await cursor.execute("select * from ora_1007")
            self.assertEqual(await cursor.fetchone(), (1, "Another String"))

    async def test_6332(self):
        "6332 - test updating an empty row"
        int_var = self.cursor.var(int)
        await self.cursor.execute("truncate table TestTempTable")
        await self.cursor.execute(
            """
            begin
                update TestTempTable set IntCol = :1
                where StringCol1 = :2
                returning IntCol into :3;
            end;
            """,
            [1, "test string 4352", int_var],
        )
        self.assertEqual(int_var.values, [None])

    async def test_6333(self):
        "6333 - fetch duplicate data from query in statement cache"
        sql = """
                select 'A', 'B', 'C' from dual
                union all
                select 'A', 'B', 'C' from dual
                union all
                select 'A', 'B', 'C' from dual"""
        expected_data = [("A", "B", "C")] * 3
        with self.conn.cursor() as cursor:
            cursor.prefetchrows = 0
            await cursor.execute(sql)
            self.assertEqual(await cursor.fetchall(), expected_data)
        with self.conn.cursor() as cursor:
            cursor.prefetchrows = 0
            await cursor.execute(sql)
            self.assertEqual(await cursor.fetchall(), expected_data)

    async def test_6334(self):
        "6334 - fetch duplicate data with outconverter"

        def out_converter(value):
            self.assertIs(type(value), str)
            return int(value)

        def type_handler(cursor, metadata):
            if metadata.name == "COL_3":
                return cursor.var(
                    str, arraysize=cursor.arraysize, outconverter=out_converter
                )

        self.cursor.outputtypehandler = type_handler
        await self.cursor.execute(
            """
            select 'A' as col_1, 2 as col_2, 3 as col_3 from dual
                union all
            select 'A' as col_1, 2 as col_2, 3 as col_3 from dual
                union all
            select 'A' as col_1, 2 as col_2, 3 as col_3 from dual
            """
        )
        expected_data = [("A", 2, 3)] * 3
        self.assertEqual(await self.cursor.fetchall(), expected_data)

    @unittest.skipIf(test_env.get_is_drcp(), "not supported with DRCP")
    async def test_6335(self):
        "6335 - kill connection with open cursor"
        admin_conn = await test_env.get_admin_connection_async()
        conn = await test_env.get_connection_async()
        self.assertEqual(conn.is_healthy(), True)
        sid, serial = await self.get_sid_serial(conn)
        with admin_conn.cursor() as admin_cursor:
            sql = f"alter system kill session '{sid},{serial}'"
            await admin_cursor.execute(sql)
        with self.assertRaisesFullCode("DPY-4011"):
            with conn.cursor() as cursor:
                await cursor.execute("select user from dual")
        self.assertFalse(conn.is_healthy())

    @unittest.skipIf(test_env.get_is_drcp(), "not supported with DRCP")
    async def test_6336(self):
        "6336 - kill connection in cursor context manager"
        admin_conn = await test_env.get_admin_connection_async()
        conn = await test_env.get_connection_async()
        self.assertEqual(conn.is_healthy(), True)
        sid, serial = await self.get_sid_serial(conn)
        with admin_conn.cursor() as admin_cursor:
            await admin_cursor.execute(
                f"alter system kill session '{sid},{serial}'"
            )
        with self.assertRaisesFullCode("DPY-4011"):
            with conn.cursor() as cursor:
                await cursor.execute("select user from dual")
        self.assertEqual(conn.is_healthy(), False)

    async def test_6337(self):
        "6337 - fetchmany() with and without parameters"
        sql_part = "select user from dual"
        sql = " union all ".join([sql_part] * 10)
        with self.conn.cursor() as cursor:
            cursor.arraysize = 6
            await cursor.execute(sql)
            rows = await cursor.fetchmany()
            self.assertEqual(len(rows), cursor.arraysize)
            await cursor.execute(sql)
            rows = await cursor.fetchmany(size=2)
            self.assertEqual(len(rows), 2)
            await cursor.execute(sql)

    async def test_6338(self):
        "6338 - access cursor.rowcount after closing cursor"
        with self.conn.cursor() as cursor:
            await cursor.execute("select user from dual")
            await cursor.fetchall()
            self.assertEqual(cursor.rowcount, 1)
        self.assertEqual(cursor.rowcount, -1)

    async def test_6339(self):
        "6339 - changing bind type with define needed"
        await self.cursor.execute("truncate table TestClobs")
        row_for_1 = (1, "Short value 1")
        row_for_56 = (56, "Short value 56")
        for data in (row_for_1, row_for_56):
            await self.cursor.execute(
                "insert into TestClobs (IntCol, ClobCol) values (:1, :2)",
                data,
            )
        sql = "select IntCol, ClobCol from TestClobs where IntCol = :int_col"
        with test_env.DefaultsContextManager("fetch_lobs", False):
            await self.cursor.execute(sql, int_col="1")
            self.assertEqual(await self.cursor.fetchone(), row_for_1)
            await self.cursor.execute(sql, int_col="56")
            self.assertEqual(await self.cursor.fetchone(), row_for_56)
            await self.cursor.execute(sql, int_col=1)
            self.assertEqual(await self.cursor.fetchone(), row_for_1)

    async def test_6340(self):
        "6340 - test calling cursor.parse() twice with the same statement"
        await self.cursor.execute("truncate table TestTempTable")
        data = (4363, "Value for test 4363")
        await self.cursor.execute(
            "insert into TestTempTable (IntCol, StringCol1) values (:1, :2)",
            data,
        )
        sql = "update TestTempTable set StringCol1 = :v where IntCol = :i"
        for i in range(2):
            await self.cursor.parse(sql)
            await self.cursor.execute(sql, ("Updated value", data[0]))

    async def test_6341(self):
        "6341 - test addition of column to cached query"
        table_name = "test_4365"
        try:
            await self.cursor.execute(f"drop table {table_name}")
        except oracledb.DatabaseError:
            pass
        data = ("val 1", "val 2")
        await self.cursor.execute(
            f"create table {table_name} (col1 varchar2(10))"
        )
        await self.cursor.execute(
            f"insert into {table_name} values (:1)", [data[0]]
        )
        await self.conn.commit()
        await self.cursor.execute(f"select * from {table_name}")
        self.assertEqual(await self.cursor.fetchall(), [(data[0],)])
        await self.cursor.execute(
            f"alter table {table_name} add col2 varchar2(10)"
        )
        await self.cursor.execute(
            f"update {table_name} set col2 = :1", [data[1]]
        )
        await self.conn.commit()
        await self.cursor.execute(f"select * from {table_name}")
        self.assertEqual(await self.cursor.fetchall(), [data])

    async def test_6342(self):
        "6342 - test executemany() with PL/SQL and increasing data lengths"
        sql = "begin :1 := length(:2); end;"
        var = self.cursor.var(int, arraysize=3)
        await self.cursor.executemany(
            sql, [(var, "one"), (var, "two"), (var, "end")]
        )
        self.assertEqual(var.values, [3, 3, 3])
        await self.cursor.executemany(
            sql, [(var, "three"), (var, "four"), (var, "end")]
        )
        self.assertEqual(var.values, [5, 4, 3])
        await self.cursor.executemany(
            sql, [(var, "five"), (var, "six"), (var, "end")]
        )
        self.assertEqual(var.values, [4, 3, 3])

    async def test_6343(self):
        "6343 - test cursor.rowcount values for queries"
        max_rows = 93
        self.cursor.arraysize = 10
        await self.cursor.execute(
            "select rownum as id from dual connect by rownum <= :1",
            [max_rows],
        )
        self.assertEqual(self.cursor.rowcount, 0)
        batch_num = 1
        while True:
            rows = await self.cursor.fetchmany()
            if not rows:
                break
            expected_value = min(max_rows, batch_num * self.cursor.arraysize)
            self.assertEqual(self.cursor.rowcount, expected_value)
            batch_num += 1
        await self.cursor.fetchall()
        self.assertEqual(self.cursor.rowcount, max_rows)

    async def test_6344(self):
        "6344 - test bind order for PL/SQL"
        await self.cursor.execute("truncate table TestClobs")
        sql = """
            insert into TestClobs (IntCol, CLOBCol, ExtraNumCol1)
            values (:1, :2, :3)"""
        data = "x" * 9000
        rows = [(1, data, 5), (2, data, 6)]
        await self.cursor.execute(sql, rows[0])
        plsql = f"begin {sql}; end;"
        await self.cursor.execute(plsql, rows[1])
        await self.conn.commit()
        with test_env.DefaultsContextManager("fetch_lobs", False):
            await self.cursor.execute(
                """
                select IntCol, CLOBCol, ExtraNumCol1
                from TestCLOBs
                order by IntCol
                """
            )
            self.assertEqual(await self.cursor.fetchall(), rows)

    async def test_6345(self):
        "6345 - test rebuild of table with LOB in cached query (as string)"
        table_name = "test_4370"
        drop_sql = f"drop table {table_name} purge"
        create_sql = f"""
            create table {table_name} (
                Col1 number(9) not null,
                Col2 clob not null
            )"""
        insert_sql = f"insert into {table_name} values (:1, :2)"
        query_sql = f"select * from {table_name} order by Col1"
        data = [(1, "CLOB value 1"), (2, "CLOB value 2")]
        try:
            await self.cursor.execute(drop_sql)
        except oracledb.DatabaseError:
            pass
        with test_env.DefaultsContextManager("fetch_lobs", False):
            await self.cursor.execute(create_sql)
            await self.cursor.executemany(insert_sql, data)
            await self.cursor.execute(query_sql)
            self.assertEqual(await self.cursor.fetchall(), data)
            await self.cursor.execute(query_sql)
            self.assertEqual(await self.cursor.fetchall(), data)
            await self.cursor.execute(drop_sql)
            await self.cursor.execute(create_sql)
            await self.cursor.executemany(insert_sql, data)
            await self.cursor.execute(query_sql)
            self.assertEqual(await self.cursor.fetchall(), data)

    async def test_6346(self):
        "6346 - test rebuild of table with LOB in cached query (as LOB)"
        table_name = "test_4371"
        drop_sql = f"drop table {table_name} purge"
        create_sql = f"""
            create table {table_name} (
                Col1 number(9) not null,
                Col2 clob not null)"""
        insert_sql = f"insert into {table_name} values (:1, :2)"
        query_sql = f"select * from {table_name} order by Col1"
        data = [(1, "CLOB value 1"), (2, "CLOB value 2")]
        try:
            await self.cursor.execute(drop_sql)
        except oracledb.DatabaseError:
            pass
        await self.cursor.execute(create_sql)
        await self.cursor.executemany(insert_sql, data)
        await self.cursor.execute(query_sql)
        fetched_data = [(n, await c.read()) async for n, c in self.cursor]
        self.assertEqual(fetched_data, data)
        await self.cursor.execute(query_sql)
        fetched_data = [(n, await c.read()) async for n, c in self.cursor]
        self.assertEqual(fetched_data, data)
        await self.cursor.execute(drop_sql)
        await self.cursor.execute(create_sql)
        await self.cursor.executemany(insert_sql, data)
        await self.cursor.execute(query_sql)
        fetched_data = [(n, await c.read()) async for n, c in self.cursor]
        self.assertEqual(fetched_data, data)

    @unittest.skipIf(
        test_env.get_server_version() < (23, 1), "unsupported database"
    )
    async def test_6347(self):
        "6347 - fetch table with domain and annotations"
        await self.cursor.execute(
            "select * from TableWithDomainAndAnnotations"
        )
        self.assertEqual(await self.cursor.fetchall(), [(1, 25)])
        column_1 = self.cursor.description[0]
        self.assertIsNone(column_1.domain_schema)
        self.assertIsNone(column_1.domain_name)
        self.assertIsNone(column_1.annotations)
        column_2 = self.cursor.description[1]
        self.assertEqual(
            column_2.domain_schema, test_env.get_main_user().upper()
        )
        self.assertEqual(column_2.domain_name, "SIMPLEDOMAIN")
        expected_annotations = {
            "ANNO_1": "first annotation",
            "ANNO_2": "second annotation",
            "ANNO_3": "",
        }
        self.assertEqual(column_2.annotations, expected_annotations)

    async def test_6348(self):
        "6348 - test fetching LOBs after an error"
        sql = """
            select
                to_clob(:val),
                1 / (dbms_lob.getlength(to_clob(:val)) - 1)
            from dual"""
        with self.assertRaisesFullCode("ORA-01476"):
            await self.cursor.execute(sql, val="a")
        await self.cursor.execute(sql, val="bb")
        lob, num_val = await self.cursor.fetchone()
        self.assertEqual(await lob.read(), "bb")
        self.assertEqual(num_val, 1)

    async def test_6349(self):
        "6349 - test parse() with autocommit enabled"
        async with test_env.get_connection_async() as conn:
            conn.autocommit = True
            cursor = conn.cursor()
            await cursor.execute("truncate table TestTempTable")
            await cursor.parse(
                "insert into TestTempTable (IntCol) values (:1)"
            )
            await cursor.execute(None, [1])

    async def test_6350(self):
        "6350 - test cursor.setinputsizes() with early failed execute"
        self.cursor.setinputsizes(a=int, b=str)
        with self.assertRaisesFullCode("DPY-2006"):
            await self.cursor.execute("select :c from dual", [5])
        value = 4368
        await self.cursor.execute("select :d from dual", [value])
        (fetched_value,) = await self.cursor.fetchone()
        self.assertEqual(fetched_value, value)

    async def test_6351(self):
        "6351 - fetch JSON columns as Python objects"
        expected_data = [
            (1, [1, 2, 3], [4, 5, 6], [7, 8, 9]),
            (2, None, None, None),
        ]
        await self.cursor.execute("select * from TestJsonCols order by IntCol")
        self.assertEqual(await self.cursor.fetchall(), expected_data)

    async def test_6352(self):
        "6352 - test fetching nested cursors repeatedly"
        sql = """
            select
                s.Description,
                cursor(select 'Nested String for ' || s.Description from dual)
            from
                (
                    select 'Top Level String 1' as Description
                    from dual
                    union all
                    select 'Top Level String 2'
                    from dual
                    union all
                    select 'Top Level String 3'
                    from dual
                    union all
                    select 'Top Level String 4'
                    from dual
                    union all
                    select 'Top Level String 5'
                    from dual
                ) s"""

        for i in range(3):
            with self.conn.cursor() as cursor:
                cursor.arraysize = 10
                await cursor.execute(sql)
                desc, nested1 = await cursor.fetchone()
                self.assertEqual(desc, "Top Level String 1")
                nested_rows = await nested1.fetchall()
                self.assertEqual(
                    nested_rows, [("Nested String for Top Level String 1",)]
                )
                desc, nested2 = await cursor.fetchone()
                self.assertEqual(desc, "Top Level String 2")
                nested_rows = await nested2.fetchall()
                self.assertEqual(
                    nested_rows, [("Nested String for Top Level String 2",)]
                )


if __name__ == "__main__":
    test_env.run_test_cases()



# -----------------------------------------------------------------------------
# Copyright (c) 2021, 2024, Oracle and/or its affiliates.
#
# This software is dual-licensed to you under the Universal Permissive License
# (UPL) 1.0 as shown at https://oss.oracle.com/licenses/upl and Apache License
# 2.0 as shown at http://www.apache.org/licenses/LICENSE-2.0. You may choose
# either license.
#
# If you elect to accept the software under the Apache License, Version 2.0,
# the following applies:
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# -----------------------------------------------------------------------------

"""
3700 - Module for testing all variable types.
"""

import datetime
import decimal
import unittest

import oracledb
import test_env


class TestCase(test_env.BaseTestCase):
    def _test_positive_set_and_get(
        self, var_type, value_to_set, expected_value, type_name=None
    ):
        var = self.cursor.var(var_type, typename=type_name)
        var.setvalue(0, value_to_set)
        result = var.getvalue()
        if isinstance(result, oracledb.LOB):
            result = result.read()
        elif isinstance(result, oracledb.DbObject):
            result = self.get_db_object_as_plain_object(result)
        if isinstance(expected_value, datetime.date) and not isinstance(
            expected_value, datetime.datetime
        ):
            if isinstance(result, datetime.datetime):
                result = result.date()
        self.assertEqual(type(result), type(expected_value))
        self.assertEqual(result, expected_value)

    def _test_negative_set_and_get(
        self, var_type, value_to_set, type_name=None
    ):
        var = self.cursor.var(var_type, typename=type_name)
        self.assertRaises(
            (TypeError, oracledb.DatabaseError), var.setvalue, 0, value_to_set
        )

    def test_3700(self):
        "3700 - setting values on variables of type DB_TYPE_NUMBER"
        self._test_positive_set_and_get(int, 5, 5)
        self._test_positive_set_and_get(oracledb.DB_TYPE_NUMBER, 3.5, 3.5)
        self._test_positive_set_and_get(
            decimal.Decimal, decimal.Decimal("24.8"), decimal.Decimal("24.8")
        )
        self._test_positive_set_and_get(int, True, 1)
        self._test_positive_set_and_get(int, False, 0)
        self._test_positive_set_and_get(oracledb.DB_TYPE_NUMBER, None, None)
        self._test_negative_set_and_get(oracledb.DB_TYPE_NUMBER, "abc")

    def test_3701(self):
        "3701 - setting values on variables of type DB_TYPE_BINARY_INTEGER"
        self._test_positive_set_and_get(oracledb.DB_TYPE_BINARY_INTEGER, 5, 5)
        self._test_positive_set_and_get(
            oracledb.DB_TYPE_BINARY_INTEGER, 3.5, 3
        )
        self._test_positive_set_and_get(
            oracledb.DB_TYPE_BINARY_INTEGER, decimal.Decimal("24.8"), 24
        )
        self._test_positive_set_and_get(
            oracledb.DB_TYPE_BINARY_INTEGER, True, 1
        )
        self._test_positive_set_and_get(
            oracledb.DB_TYPE_BINARY_INTEGER, False, 0
        )
        self._test_positive_set_and_get(
            oracledb.DB_TYPE_BINARY_INTEGER, None, None
        )
        self._test_negative_set_and_get(oracledb.DB_TYPE_BINARY_INTEGER, "abc")

    def test_3702(self):
        "3702 - setting values on variables of type DB_TYPE_VARCHAR"
        value = "A VARCHAR string"
        self._test_positive_set_and_get(oracledb.DB_TYPE_VARCHAR, value, value)
        value = b"A raw string for VARCHAR"
        self._test_positive_set_and_get(
            oracledb.DB_TYPE_VARCHAR, value, value.decode()
        )
        self._test_positive_set_and_get(oracledb.DB_TYPE_VARCHAR, None, None)
        self._test_negative_set_and_get(oracledb.DB_TYPE_VARCHAR, 5)

    def test_3703(self):
        "3703 - setting values on variables of type DB_TYPE_NVARCHAR"
        value = "A NVARCHAR string"
        self._test_positive_set_and_get(
            oracledb.DB_TYPE_NVARCHAR, value, value
        )
        value = b"A raw string for NVARCHAR"
        self._test_positive_set_and_get(
            oracledb.DB_TYPE_NVARCHAR, value, value.decode()
        )
        self._test_positive_set_and_get(oracledb.DB_TYPE_NVARCHAR, None, None)
        self._test_negative_set_and_get(oracledb.DB_TYPE_NVARCHAR, 5)

    def test_3704(self):
        "3704 - setting values on variables of type DB_TYPE_CHAR"
        value = "A CHAR string"
        self._test_positive_set_and_get(oracledb.DB_TYPE_CHAR, value, value)
        value = b"A raw string for CHAR"
        self._test_positive_set_and_get(
            oracledb.DB_TYPE_CHAR, value, value.decode()
        )
        self._test_positive_set_and_get(oracledb.DB_TYPE_CHAR, None, None)
        self._test_negative_set_and_get(oracledb.DB_TYPE_CHAR, 5)

    def test_3705(self):
        "3705 - setting values on variables of type DB_TYPE_NCHAR"
        value = "A NCHAR string"
        self._test_positive_set_and_get(oracledb.DB_TYPE_NCHAR, value, value)
        value = b"A raw string for NCHAR"
        self._test_positive_set_and_get(
            oracledb.DB_TYPE_CHAR, value, value.decode()
        )
        self._test_positive_set_and_get(oracledb.DB_TYPE_NCHAR, None, None)
        self._test_negative_set_and_get(oracledb.DB_TYPE_NCHAR, 5)

    def test_3706(self):
        "3706 - setting values on variables of type DB_TYPE_LONG"
        value = "Long Data" * 15000
        self._test_positive_set_and_get(oracledb.DB_TYPE_LONG, value, value)
        value = b"Raw data for LONG" * 15000
        self._test_positive_set_and_get(
            oracledb.DB_TYPE_LONG, value, value.decode()
        )
        self._test_positive_set_and_get(oracledb.DB_TYPE_LONG, None, None)
        self._test_negative_set_and_get(oracledb.DB_TYPE_LONG, 5)

    def test_3707(self):
        "3707 - setting values on variables of type DB_TYPE_RAW"
        value = b"Raw Data"
        self._test_positive_set_and_get(oracledb.DB_TYPE_RAW, value, value)
        value = "String data for RAW"
        self._test_positive_set_and_get(
            oracledb.DB_TYPE_RAW, value, value.encode()
        )
        self._test_positive_set_and_get(oracledb.DB_TYPE_RAW, None, None)
        self._test_negative_set_and_get(oracledb.DB_TYPE_RAW, 5)

    def test_3708(self):
        "3708 - setting values on variables of type DB_TYPE_LONG_RAW"
        value = b"Long Raw Data" * 15000
        self._test_positive_set_and_get(
            oracledb.DB_TYPE_LONG_RAW, value, value
        )
        value = "String data for LONG RAW" * 15000
        self._test_positive_set_and_get(
            oracledb.DB_TYPE_LONG_RAW, value, value.encode()
        )
        self._test_positive_set_and_get(oracledb.DB_TYPE_LONG_RAW, None, None)
        self._test_negative_set_and_get(oracledb.DB_TYPE_LONG_RAW, 5)

    def test_3709(self):
        "3709 - setting values on variables of type DB_TYPE_DATE"
        value = datetime.date(2017, 5, 6)
        self._test_positive_set_and_get(oracledb.DB_TYPE_DATE, value, value)
        value = datetime.datetime(2017, 5, 6, 9, 36, 0)
        self._test_positive_set_and_get(oracledb.DB_TYPE_DATE, value, value)
        self._test_positive_set_and_get(oracledb.DB_TYPE_DATE, None, None)
        self._test_negative_set_and_get(oracledb.DB_TYPE_DATE, 5)

    def test_3710(self):
        "3710 - setting values on variables of type DB_TYPE_TIMESTAMP"
        value = datetime.date(2017, 5, 6)
        self._test_positive_set_and_get(
            oracledb.DB_TYPE_TIMESTAMP, value, value
        )
        value = datetime.datetime(2017, 5, 6, 9, 36, 0, 300000)
        self._test_positive_set_and_get(
            oracledb.DB_TYPE_TIMESTAMP, value, value
        )
        self._test_positive_set_and_get(oracledb.DB_TYPE_TIMESTAMP, None, None)
        self._test_negative_set_and_get(oracledb.DB_TYPE_TIMESTAMP, 5)

    def test_3711(self):
        "3711 - setting values on variables of type DB_TYPE_TIMESTAMP_TZ"
        value = datetime.date(2017, 5, 6)
        self._test_positive_set_and_get(
            oracledb.DB_TYPE_TIMESTAMP_TZ, value, value
        )
        value = datetime.datetime(2017, 5, 6, 9, 36, 0, 300000)
        self._test_positive_set_and_get(
            oracledb.DB_TYPE_TIMESTAMP_TZ, value, value
        )
        self._test_positive_set_and_get(
            oracledb.DB_TYPE_TIMESTAMP_TZ, None, None
        )
        self._test_negative_set_and_get(oracledb.DB_TYPE_TIMESTAMP_TZ, 5)

    def test_3712(self):
        "3712 - setting values on variables of type DB_TYPE_TIMESTAMP_LTZ"
        value = datetime.date(2017, 5, 6)
        self._test_positive_set_and_get(
            oracledb.DB_TYPE_TIMESTAMP_LTZ, value, value
        )
        value = datetime.datetime(2017, 5, 6, 9, 36, 0, 300000)
        self._test_positive_set_and_get(
            oracledb.DB_TYPE_TIMESTAMP_LTZ, value, value
        )
        self._test_positive_set_and_get(
            oracledb.DB_TYPE_TIMESTAMP_LTZ, None, None
        )
        self._test_negative_set_and_get(oracledb.DB_TYPE_TIMESTAMP_LTZ, 5)

    def test_3713(self):
        "3713 - setting values on variables of type DB_TYPE_BLOB"
        value = b"Short temp BLOB value"
        temp_blob = self.conn.createlob(oracledb.DB_TYPE_BLOB)
        temp_blob.write(value)
        self._test_positive_set_and_get(
            oracledb.DB_TYPE_BLOB, temp_blob, value
        )
        self._test_negative_set_and_get(oracledb.DB_TYPE_CLOB, temp_blob)
        self._test_negative_set_and_get(oracledb.DB_TYPE_NCLOB, temp_blob)
        value = b"Short BLOB value"
        self._test_positive_set_and_get(oracledb.DB_TYPE_BLOB, value, value)
        self._test_positive_set_and_get(oracledb.DB_TYPE_BLOB, None, None)
        self._test_negative_set_and_get(oracledb.DB_TYPE_BLOB, 5)

    def test_3714(self):
        "3714 - setting values on variables of type DB_TYPE_CLOB"
        value = "Short temp CLOB value"
        temp_clob = self.conn.createlob(oracledb.DB_TYPE_CLOB)
        temp_clob.write(value)
        self._test_positive_set_and_get(
            oracledb.DB_TYPE_CLOB, temp_clob, value
        )
        self._test_negative_set_and_get(oracledb.DB_TYPE_BLOB, temp_clob)
        self._test_negative_set_and_get(oracledb.DB_TYPE_NCLOB, temp_clob)
        value = "Short CLOB value"
        self._test_positive_set_and_get(oracledb.DB_TYPE_CLOB, value, value)
        self._test_positive_set_and_get(oracledb.DB_TYPE_CLOB, None, None)
        self._test_negative_set_and_get(oracledb.DB_TYPE_CLOB, 5)

    def test_3715(self):
        "3715 - setting values on variables of type DB_TYPE_NCLOB"
        value = "Short temp NCLOB value"
        temp_nclob = self.conn.createlob(oracledb.DB_TYPE_NCLOB)
        temp_nclob.write(value)
        self._test_positive_set_and_get(
            oracledb.DB_TYPE_NCLOB, temp_nclob, value
        )
        self._test_negative_set_and_get(oracledb.DB_TYPE_BLOB, temp_nclob)
        self._test_negative_set_and_get(oracledb.DB_TYPE_CLOB, temp_nclob)
        value = "Short NCLOB Value"
        self._test_positive_set_and_get(oracledb.DB_TYPE_NCLOB, value, value)
        self._test_positive_set_and_get(oracledb.DB_TYPE_NCLOB, None, None)
        self._test_negative_set_and_get(oracledb.DB_TYPE_NCLOB, 5)

    def test_3716(self):
        "3716 - setting values on variables of type DB_TYPE_BINARY_FLOAT"
        self._test_positive_set_and_get(oracledb.DB_TYPE_BINARY_FLOAT, 5, 5.0)
        self._test_positive_set_and_get(
            oracledb.DB_TYPE_BINARY_FLOAT, 3.5, 3.5
        )
        self._test_positive_set_and_get(
            oracledb.DB_TYPE_BINARY_FLOAT, decimal.Decimal("24.5"), 24.5
        )
        self._test_positive_set_and_get(
            oracledb.DB_TYPE_BINARY_FLOAT, True, 1.0
        )
        self._test_positive_set_and_get(
            oracledb.DB_TYPE_BINARY_FLOAT, False, 0.0
        )
        self._test_positive_set_and_get(
            oracledb.DB_TYPE_BINARY_FLOAT, None, None
        )
        self._test_negative_set_and_get(oracledb.DB_TYPE_BINARY_FLOAT, "abc")

    def test_3717(self):
        "3717 - setting values on variables of type DB_TYPE_BINARY_DOUBLE"
        self._test_positive_set_and_get(oracledb.DB_TYPE_BINARY_DOUBLE, 5, 5.0)
        self._test_positive_set_and_get(
            oracledb.DB_TYPE_BINARY_DOUBLE, 3.5, 3.5
        )
        self._test_positive_set_and_get(
            oracledb.DB_TYPE_BINARY_DOUBLE, decimal.Decimal("192.125"), 192.125
        )
        self._test_positive_set_and_get(
            oracledb.DB_TYPE_BINARY_DOUBLE, True, 1.0
        )
        self._test_positive_set_and_get(
            oracledb.DB_TYPE_BINARY_DOUBLE, False, 0.0
        )
        self._test_positive_set_and_get(
            oracledb.DB_TYPE_BINARY_DOUBLE, None, None
        )
        self._test_negative_set_and_get(oracledb.DB_TYPE_BINARY_DOUBLE, "abc")

    def test_3718(self):
        "3718 - setting values on variables of type DB_TYPE_BOOLEAN"
        self._test_positive_set_and_get(oracledb.DB_TYPE_BOOLEAN, 5, True)
        self._test_positive_set_and_get(oracledb.DB_TYPE_BOOLEAN, 2.0, True)
        self._test_positive_set_and_get(oracledb.DB_TYPE_BOOLEAN, "abc", True)
        self._test_positive_set_and_get(
            oracledb.DB_TYPE_BOOLEAN, decimal.Decimal("24.8"), True
        )
        self._test_positive_set_and_get(oracledb.DB_TYPE_BOOLEAN, 0.0, False)
        self._test_positive_set_and_get(oracledb.DB_TYPE_BOOLEAN, 0, False)
        self._test_positive_set_and_get(oracledb.DB_TYPE_BOOLEAN, None, None)

    def test_3719(self):
        "3719 - setting values on variables of type DB_TYPE_INTERVAL_DS"
        value = datetime.timedelta(days=5, seconds=56000, microseconds=123780)
        self._test_positive_set_and_get(
            oracledb.DB_TYPE_INTERVAL_DS, value, value
        )
        self._test_positive_set_and_get(
            oracledb.DB_TYPE_INTERVAL_DS, None, None
        )
        self._test_negative_set_and_get(oracledb.DB_TYPE_INTERVAL_DS, 5)

    def test_3720(self):
        "3720 - setting values on variables of type DB_TYPE_ROWID"
        self._test_negative_set_and_get(oracledb.DB_TYPE_ROWID, 12345)
        self._test_negative_set_and_get(oracledb.DB_TYPE_ROWID, "523lkhlf")

    def test_3721(self):
        "3721 - setting values on variables of type DB_TYPE_OBJECT"
        obj_type = self.conn.gettype("UDT_OBJECT")
        obj = obj_type.newobject()
        plain_obj = self.get_db_object_as_plain_object(obj)
        self._test_positive_set_and_get(
            oracledb.DB_TYPE_OBJECT, obj, plain_obj, "UDT_OBJECT"
        )
        self._test_positive_set_and_get(obj_type, obj, plain_obj)
        self._test_positive_set_and_get(
            oracledb.DB_TYPE_OBJECT, None, None, "UDT_OBJECT"
        )
        self._test_positive_set_and_get(obj_type, None, None)
        self._test_negative_set_and_get(
            oracledb.DB_TYPE_OBJECT, "abc", "UDT_OBJECT"
        )
        self._test_negative_set_and_get(
            oracledb.DB_TYPE_OBJECT, obj, "UDT_OBJECTARRAY"
        )
        wrong_obj_type = self.conn.gettype("UDT_OBJECTARRAY")
        self._test_negative_set_and_get(wrong_obj_type, obj)

    @unittest.skipIf(
        test_env.get_client_version() < (21, 0), "unsupported client"
    )
    @unittest.skipIf(
        test_env.get_server_version() < (21, 0), "unsupported server"
    )
    def test_3722(self):
        "3722 - setting values on variables of type DB_TYPE_JSON"
        json_data = [
            5,
            25.25,
            decimal.Decimal("10.25"),
            True,
            False,
            datetime.datetime(2017, 5, 6),
            datetime.datetime(2017, 5, 6, 9, 36, 0, 300000),
            datetime.timedelta(days=5, seconds=56000, microseconds=123780),
            {},
            "String",
            b"Some bytes",
            {"keyA": 1, "KeyB": "Melbourne"},
            [],
            [1, "A"],
            {"name": None},
            {"name": "John"},
            {"age": 30},
            {"Permanent": True},
            {
                "employee": {
                    "name": "John",
                    "age": 30,
                    "city": "Delhi",
                    "Parmanent": True,
                }
            },
            {"employees": ["John", "Matthew", "James"]},
            {
                "employees": [
                    {"employee1": {"name": "John", "city": "Delhi"}},
                    {"employee2": {"name": "Matthew", "city": "Mumbai"}},
                    {"employee3": {"name": "James", "city": "Bangalore"}},
                ]
            },
        ]
        self._test_positive_set_and_get(
            oracledb.DB_TYPE_JSON, json_data, json_data
        )
        self._test_positive_set_and_get(oracledb.DB_TYPE_JSON, None, None)

    def test_3723(self):
        "3723 - test setting values on variables of type DB_TYPE_CURSOR"
        self._test_positive_set_and_get(oracledb.DB_TYPE_CURSOR, None, None)
        self._test_negative_set_and_get(oracledb.DB_TYPE_CURSOR, 5)

    def test_3724(self):
        "3724 - test fetching columns containing all null values"
        self.cursor.execute(
            """
            select null, to_char(null), to_number(null), to_date(null),
                to_timestamp(null), to_clob(null), to_blob(null)
            from dual
            """
        )
        self.assertEqual(
            self.cursor.fetchall(),
            [(None, None, None, None, None, None, None)],
        )

    @unittest.skipIf(
        not test_env.get_is_thin(), "thick mode doesn't support DB_TYPE_UROWID"
    )
    def test_3725(self):
        "3725 - setting values on variables of type DB_TYPE_UROWID"
        self._test_negative_set_and_get(oracledb.DB_TYPE_UROWID, 12345)
        self._test_negative_set_and_get(oracledb.DB_TYPE_UROWID, "523lkhlf")

    def test_3726(self):
        "3726 - getting value with an specific index"
        var = self.cursor.var(oracledb.DB_TYPE_NUMBER, 1000, 2)
        var.setvalue(0, 10)
        self.assertEqual(var.getvalue(0), 10)
        self.assertIsNone(var.getvalue(1))
        self.assertRaises(IndexError, var.getvalue, 4)

    def test_3727(self):
        "3727 - getting buffer_size attribute"
        test_values = [
            (oracledb.DB_TYPE_NUMBER, 200, 22),
            (oracledb.DB_TYPE_VARCHAR, 3000, 12000),
            (oracledb.DB_TYPE_RAW, 4000, 4000),
            (oracledb.DB_TYPE_NCHAR, 1000, 4000),
            (oracledb.DB_TYPE_CHAR, 2000, 8000),
        ]
        for typ, size, buffer_size in test_values:
            var = self.cursor.var(typ, size)
            self.assertEqual(var.buffer_size, buffer_size)

    def test_3728(self):
        "3728 - getting actual elements"
        array_size = 8
        var = self.cursor.var(oracledb.DB_TYPE_NUMBER, arraysize=array_size)
        self.assertEqual(var.actual_elements, array_size)
        self.assertEqual(var.actual_elements, var.num_elements)

    def test_3729(self):
        "3729 - test deprecated attributes"
        var = self.cursor.var(oracledb.DB_TYPE_NUMBER, arraysize=200)
        self.assertEqual(var.bufferSize, 22)
        self.assertEqual(var.actualElements, 200)
        self.assertEqual(var.numElements, 200)

    def test_3730(self):
        "3730 - test calling of outconverter with null values"

        def type_handler(cursor, metadata):
            return cursor.var(
                metadata.type_code,
                outconverter=lambda v: f"|{v}|" if v else "",
                convert_nulls=True,
                arraysize=cursor.arraysize,
            )

        self.cursor.outputtypehandler = type_handler
        self.cursor.execute(
            """
            select 'First - A', 'First - B'
            from dual
                union all
            select 'Second - A', null
            from dual
                union all
            select null, 'Third - B'
            from dual
            """
        )
        rows = self.cursor.fetchall()
        expected_rows = [
            ("|First - A|", "|First - B|"),
            ("|Second - A|", ""),
            ("", "|Third - B|"),
        ]
        self.assertEqual(rows, expected_rows)

    def test_3731(self):
        "3731 - test getting convert_nulls"
        for convert_nulls in [True, False]:
            simple_var = self.cursor.var(str, convert_nulls=convert_nulls)
            self.assertEqual(simple_var.convert_nulls, convert_nulls)

    def test_3732(self):
        "3732 - test encoding_errors"
        if test_env.get_charset() != "AL32UTF8":
            self.skipTest("Database character set must be AL32UTF8")
        str_value = "Я"
        replacement_char = "�"
        invalid_bytes = str_value.encode("windows-1251")
        self.cursor.execute("truncate table TestTempTable")
        self.cursor.execute(
            """
            insert into TestTempTable (IntCol, StringCol1)
            values (1, utl_raw.cast_to_varchar2(:1))
            """,
            [invalid_bytes],
        )
        self.conn.commit()

        for arg_name in ["encoding_errors", "encodingErrors"]:
            with self.subTest(arg_name=arg_name):

                def type_handler(cursor, fetch_info):
                    args = dict(arraysize=cursor.arraysize)
                    args[arg_name] = "replace"
                    return cursor.var(fetch_info.type_code, **args)

                with self.conn.cursor() as cursor:
                    cursor.outputtypehandler = type_handler
                    cursor.execute("select StringCol1 from TestTempTable")
                    (fetched_value,) = cursor.fetchone()
                    self.assertEqual(fetched_value, replacement_char)


if __name__ == "__main__":
    test_env.run_test_cases()



# -----------------------------------------------------------------------------
# Copyright (c) 2025, Oracle and/or its affiliates.
#
# This software is dual-licensed to you under the Universal Permissive License
# (UPL) 1.0 as shown at https://oss.oracle.com/licenses/upl and Apache License
# 2.0 as shown at http://www.apache.org/licenses/LICENSE-2.0. You may choose
# either license.
#
# If you elect to accept the software under the Apache License, Version 2.0,
# the following applies:
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# -----------------------------------------------------------------------------

"""
8400 - Module for testing AQ with DbObject payloads with asyncio.
"""

import decimal
import unittest

import oracledb
import test_env


@unittest.skipUnless(
    test_env.get_is_thin(), "asyncio not supported in thick mode"
)
class TestCase(test_env.BaseAsyncTestCase):
    book_type_name = "UDT_BOOK"
    book_queue_name = "TEST_BOOK_QUEUE"
    book_data = [
        ("Wings of Fire", "A.P.J. Abdul Kalam", decimal.Decimal("15.75")),
        ("The Story of My Life", "Hellen Keller", decimal.Decimal("10.50")),
        ("The Chronicles of Narnia", "C.S. Lewis", decimal.Decimal("25.25")),
    ]

    def __verify_attr(self, obj, attrName, value):
        setattr(obj, attrName, value)
        self.assertEqual(getattr(obj, attrName), value)

    async def test_8400(self):
        "8400 - test dequeuing an empty queue"
        queue = await self.get_and_clear_queue(
            self.book_queue_name, self.book_type_name
        )
        queue.deqoptions.wait = oracledb.DEQ_NO_WAIT
        props = await queue.deqone()
        self.assertIsNone(props)

    async def test_8401(self):
        "8401 - test enqueuing and dequeuing multiple messages"
        queue = await self.get_and_clear_queue(
            self.book_queue_name, self.book_type_name
        )
        props = self.conn.msgproperties()
        for title, authors, price in self.book_data:
            props.payload = book = queue.payload_type.newobject()
            book.TITLE = title
            book.AUTHORS = authors
            book.PRICE = price
            await queue.enqone(props)
        queue.deqoptions.navigation = oracledb.DEQ_FIRST_MSG
        queue.deqoptions.wait = oracledb.DEQ_NO_WAIT
        results = []
        while True:
            props = await queue.deqone()
            if props is None:
                break
            book = props.payload
            row = (book.TITLE, book.AUTHORS, book.PRICE)
            results.append(row)
        await self.conn.commit()
        self.assertEqual(results, self.book_data)

    async def test_8402(self):
        "8402 - test dequeuing with DEQ_REMOVE_NODATA option"
        queue = await self.get_and_clear_queue(
            self.book_queue_name, self.book_type_name
        )
        book = queue.payload_type.newobject()
        book.TITLE, book.AUTHORS, book.PRICE = self.book_data[1]
        props = self.conn.msgproperties(payload=book)
        await queue.enqone(props)
        queue.deqoptions.navigation = oracledb.DEQ_FIRST_MSG
        queue.deqoptions.wait = oracledb.DEQ_NO_WAIT
        queue.deqoptions.mode = oracledb.DEQ_REMOVE_NODATA
        props = await queue.deqone()
        self.assertIsNotNone(props)
        self.assertIsNone(props.payload.TITLE)

    async def test_8403(self):
        "8403 - test getting/setting dequeue options attributes"
        queue = await self.get_and_clear_queue(
            self.book_queue_name, self.book_type_name
        )
        options = queue.deqoptions
        self.__verify_attr(options, "condition", "TEST_CONDITION")
        self.__verify_attr(options, "consumername", "TEST_CONSUMERNAME")
        self.__verify_attr(options, "correlation", "TEST_CORRELATION")
        self.__verify_attr(options, "mode", oracledb.DEQ_LOCKED)
        self.__verify_attr(
            options, "navigation", oracledb.DEQ_NEXT_TRANSACTION
        )
        self.__verify_attr(options, "transformation", "TEST_TRANSFORMATION")
        self.__verify_attr(options, "visibility", oracledb.ENQ_IMMEDIATE)
        self.__verify_attr(options, "wait", 1287)
        self.__verify_attr(options, "msgid", b"mID")

    async def test_8404(self):
        "8404 - test getting/setting enqueue options attributes"
        queue = await self.get_and_clear_queue(
            self.book_queue_name, self.book_type_name
        )
        options = queue.enqoptions
        self.__verify_attr(options, "visibility", oracledb.ENQ_IMMEDIATE)

    async def test_8405(self):
        "8405 - test errors for invalid values for enqueue"
        queue = await self.get_and_clear_queue(
            self.book_queue_name, self.book_type_name
        )
        book = queue.payload_type.newobject()
        with self.assertRaises(TypeError):
            await queue.enqone(book)

    async def test_8406(self):
        "8406 - test getting/setting message properties attributes"
        props = self.conn.msgproperties()
        self.__verify_attr(props, "correlation", "TEST_CORRELATION")
        self.__verify_attr(props, "delay", 60)
        self.__verify_attr(props, "exceptionq", "TEST_EXCEPTIONQ")
        self.__verify_attr(props, "expiration", 30)
        self.assertEqual(props.attempts, 0)
        self.__verify_attr(props, "priority", 1)
        self.assertEqual(props.state, oracledb.MSG_READY)
        self.assertEqual(props.deliverymode, 0)

    async def test_8407(self):
        "8407 - test enqueue visibility option - ENQ_ON_COMMIT"
        queue = await self.get_and_clear_queue(
            self.book_queue_name, self.book_type_name
        )
        book = queue.payload_type.newobject()
        book.TITLE, book.AUTHORS, book.PRICE = self.book_data[0]
        queue.enqoptions.visibility = oracledb.ENQ_ON_COMMIT
        props = self.conn.msgproperties(payload=book)
        await queue.enqone(props)

        async with test_env.get_connection_async() as other_conn:
            books_type = await other_conn.gettype(self.book_type_name)
            queue = other_conn.queue(self.book_queue_name, books_type)
            queue.deqoptions.navigation = oracledb.DEQ_FIRST_MSG
            queue.deqoptions.wait = oracledb.DEQ_NO_WAIT
            props = await queue.deqone()
            self.assertIsNone(props)
            await self.conn.commit()
            props = await queue.deqone()
            await other_conn.commit()
            self.assertIsNotNone(props)

    async def test_8408(self):
        "8408 - test enqueue visibility option - ENQ_IMMEDIATE"
        queue = await self.get_and_clear_queue(
            self.book_queue_name, self.book_type_name
        )
        book = queue.payload_type.newobject()
        book.TITLE, book.AUTHORS, book.PRICE = self.book_data[0]
        queue.enqoptions.visibility = oracledb.ENQ_IMMEDIATE
        props = self.conn.msgproperties(payload=book)
        await queue.enqone(props)

        async with test_env.get_connection_async() as other_conn:
            books_type = await other_conn.gettype(self.book_type_name)
            queue = other_conn.queue(self.book_queue_name, books_type)
            queue.deqoptions.navigation = oracledb.DEQ_FIRST_MSG
            queue.deqoptions.visibility = oracledb.DEQ_ON_COMMIT
            queue.deqoptions.wait = oracledb.DEQ_NO_WAIT
            props = await queue.deqone()
            book = props.payload
            results = (book.TITLE, book.AUTHORS, book.PRICE)
            await other_conn.commit()
            self.assertEqual(results, self.book_data[0])

    async def test_8409(self):
        "8409 - test enqueue/dequeue delivery modes identical - buffered"
        queue = await self.get_and_clear_queue(
            self.book_queue_name, self.book_type_name
        )
        book = queue.payload_type.newobject()
        book.TITLE, book.AUTHORS, book.PRICE = self.book_data[0]
        queue.enqoptions.deliverymode = oracledb.MSG_BUFFERED
        queue.enqoptions.visibility = oracledb.ENQ_IMMEDIATE
        props = self.conn.msgproperties(payload=book)
        await queue.enqone(props)

        async with test_env.get_connection_async() as other_conn:
            books_type = await other_conn.gettype(self.book_type_name)
            queue = other_conn.queue(self.book_queue_name, books_type)
            queue.deqoptions.deliverymode = oracledb.MSG_BUFFERED
            queue.deqoptions.navigation = oracledb.DEQ_FIRST_MSG
            queue.deqoptions.visibility = oracledb.DEQ_IMMEDIATE
            queue.deqoptions.wait = oracledb.DEQ_NO_WAIT
            props = await queue.deqone()
            book = props.payload
            results = (book.TITLE, book.AUTHORS, book.PRICE)
            await other_conn.commit()
            self.assertEqual(results, self.book_data[0])

    async def test_8410(self):
        "8410 - test enqueue/dequeue delivery modes identical - persistent"
        queue = await self.get_and_clear_queue(
            self.book_queue_name, self.book_type_name
        )
        book = queue.payload_type.newobject()
        book.TITLE, book.AUTHORS, book.PRICE = self.book_data[0]
        queue.enqoptions.deliverymode = oracledb.MSG_PERSISTENT
        queue.enqoptions.visibility = oracledb.ENQ_IMMEDIATE
        props = self.conn.msgproperties(payload=book)
        await queue.enqone(props)

        async with test_env.get_connection_async() as other_conn:
            books_type = await other_conn.gettype(self.book_type_name)
            queue = other_conn.queue(self.book_queue_name, books_type)
            queue.deqoptions.deliverymode = oracledb.MSG_PERSISTENT
            queue.deqoptions.navigation = oracledb.DEQ_FIRST_MSG
            queue.deqoptions.visibility = oracledb.DEQ_IMMEDIATE
            queue.deqoptions.wait = oracledb.DEQ_NO_WAIT
            props = await queue.deqone()
            book = props.payload
            results = (book.TITLE, book.AUTHORS, book.PRICE)
            await other_conn.commit()
            self.assertEqual(results, self.book_data[0])

    async def test_8411(self):
        "8411 - test enqueue/dequeue delivery modes the same"
        queue = await self.get_and_clear_queue(
            self.book_queue_name, self.book_type_name
        )
        book = queue.payload_type.newobject()
        book.TITLE, book.AUTHORS, book.PRICE = self.book_data[0]
        queue.enqoptions.deliverymode = oracledb.MSG_PERSISTENT_OR_BUFFERED
        queue.enqoptions.visibility = oracledb.ENQ_IMMEDIATE
        props = self.conn.msgproperties(payload=book)
        await queue.enqone(props)

        async with test_env.get_connection_async() as other_conn:
            books_type = await other_conn.gettype(self.book_type_name)
            queue = other_conn.queue(self.book_queue_name, books_type)
            queue.deqoptions.deliverymode = oracledb.MSG_PERSISTENT_OR_BUFFERED
            queue.deqoptions.navigation = oracledb.DEQ_FIRST_MSG
            queue.deqoptions.visibility = oracledb.DEQ_IMMEDIATE
            queue.deqoptions.wait = oracledb.DEQ_NO_WAIT
            props = await queue.deqone()
            book = props.payload
            results = (book.TITLE, book.AUTHORS, book.PRICE)
            await other_conn.commit()
            self.assertEqual(results, self.book_data[0])

    async def test_8412(self):
        "8412 - test enqueue/dequeue delivery modes different"
        queue = await self.get_and_clear_queue(
            self.book_queue_name, self.book_type_name
        )
        book = queue.payload_type.newobject()
        book.TITLE, book.AUTHORS, book.PRICE = self.book_data[0]
        queue.enqoptions.deliverymode = oracledb.MSG_BUFFERED
        queue.enqoptions.visibility = oracledb.ENQ_IMMEDIATE
        props = self.conn.msgproperties(payload=book)
        await queue.enqone(props)

        async with test_env.get_connection_async() as other_conn:
            books_type = await other_conn.gettype(self.book_type_name)
            queue = other_conn.queue(self.book_queue_name, books_type)
            queue.deqoptions.deliverymode = oracledb.MSG_PERSISTENT
            queue.deqoptions.navigation = oracledb.DEQ_FIRST_MSG
            queue.deqoptions.visibility = oracledb.DEQ_IMMEDIATE
            queue.deqoptions.wait = oracledb.DEQ_NO_WAIT
            props = await queue.deqone()
            self.assertIsNone(props)

    async def test_8413(self):
        "8413 - test error for message with no payload"
        books_type = await self.conn.gettype(self.book_type_name)
        queue = self.conn.queue(self.book_queue_name, books_type)
        props = self.conn.msgproperties()
        with self.assertRaisesFullCode("DPY-2000"):
            await queue.enqone(props)

    async def test_8414(self):
        "8414 - verify that the msgid property is returned correctly"
        queue = await self.get_and_clear_queue(
            self.book_queue_name, self.book_type_name
        )
        book = queue.payload_type.newobject()
        book.TITLE, book.AUTHORS, book.PRICE = self.book_data[0]
        props = self.conn.msgproperties(payload=book)
        self.assertIsNone(props.msgid)
        await queue.enqone(props)
        await self.cursor.execute("select msgid from book_queue_tab")
        (actual_msgid,) = await self.cursor.fetchone()
        self.assertEqual(props.msgid, actual_msgid)
        props = await queue.deqone()
        self.assertEqual(props.msgid, actual_msgid)

    async def test_8415(self):
        "8415 - test message props enqtime"
        queue = await self.get_and_clear_queue(
            self.book_queue_name, self.book_type_name
        )
        book = queue.payload_type.newobject()
        await self.cursor.execute("select current_timestamp from dual")
        (start_date,) = await self.cursor.fetchone()
        start_date = start_date.replace(microsecond=0)
        props = self.conn.msgproperties(payload=book)
        await queue.enqone(props)
        props = await queue.deqone()
        await self.cursor.execute("select current_timestamp from dual")
        (end_date,) = await self.cursor.fetchone()
        end_date = end_date.replace(microsecond=0)
        self.assertTrue(start_date <= props.enqtime <= end_date)

    async def test_8416(self):
        "8416 - test message props declared attributes"
        queue = await self.get_and_clear_queue(
            self.book_queue_name, self.book_type_name
        )
        book = queue.payload_type.newobject()
        values = dict(
            payload=book,
            correlation="TEST_CORRELATION",
            delay=7,
            exceptionq="TEST_EXCEPTIONQ",
            expiration=10,
            priority=1,
        )
        props = self.conn.msgproperties(**values)
        for attr_name in values:
            self.assertEqual(getattr(props, attr_name), values[attr_name])

    async def test_8417(self):
        "8417 - test error for invalid type for payload_type"
        self.assertRaises(
            TypeError, self.conn.queue, "THE QUEUE", payload_type=4
        )

    async def test_8418(self):
        "8418 - test getting queue attributes"
        queue = await self.get_and_clear_queue(
            self.book_queue_name, self.book_type_name
        )
        self.assertEqual(queue.name, self.book_queue_name)
        self.assertEqual(queue.connection, self.conn)

    async def test_8419(self):
        "8419 - test getting write-only attributes"
        queue = await self.get_and_clear_queue(
            self.book_queue_name, self.book_type_name
        )
        with self.assertRaises(AttributeError):
            queue.enqoptions.deliverymode
        with self.assertRaises(AttributeError):
            queue.deqoptions.deliverymode

    async def test_8420(self):
        "8420 - test correlation deqoption"
        queue = await self.get_and_clear_queue(
            self.book_queue_name, self.book_type_name
        )
        book = queue.payload_type.newobject()
        book.TITLE, book.AUTHORS, book.PRICE = self.book_data[0]
        correlations = ["Math", "Programming"]
        num_messages = 3
        messages = [
            self.conn.msgproperties(payload=book, correlation=c)
            for c in correlations
            for i in range(num_messages)
        ]
        await queue.enqmany(messages)
        queue.deqoptions.wait = oracledb.DEQ_NO_WAIT
        queue.deqoptions.correlation = correlations[0]
        correlated_messages = await queue.deqmany(num_messages + 1)
        self.assertEqual(len(correlated_messages), num_messages)

        queue.deqoptions.correlation = correlations[1]
        with self.assertRaisesFullCode("ORA-25241"):
            await queue.deqone()
        queue.deqoptions.navigation = oracledb.DEQ_FIRST_MSG
        correlated_messages = await queue.deqmany(num_messages + 1)
        self.assertEqual(len(correlated_messages), num_messages)

    async def test_8421(self):
        "8421 - test correlation deqoption with pattern-matching characters"
        queue = await self.get_and_clear_queue(
            self.book_queue_name, self.book_type_name
        )
        book = queue.payload_type.newobject()
        book.TITLE, book.AUTHORS, book.PRICE = self.book_data[0]
        for correlation in ("PreCalculus-math1", "Calculus-Math2"):
            props = self.conn.msgproperties(
                payload=book, correlation=correlation
            )
            await queue.enqone(props)
        queue.deqoptions.wait = oracledb.DEQ_NO_WAIT
        queue.deqoptions.correlation = "%Calculus-%ath_"
        messages = await queue.deqmany(5)
        self.assertEqual(len(messages), 2)

    async def test_8422(self):
        "8422 - test condition deqoption with priority"
        queue = await self.get_and_clear_queue(
            self.book_queue_name, self.book_type_name
        )
        queue.deqoptions.wait = oracledb.DEQ_NO_WAIT

        priorities = [5, 10]
        indexes = [0, 1]
        for priority, ix in zip(priorities, indexes):
            book = queue.payload_type.newobject()
            book.TITLE, book.AUTHORS, book.PRICE = self.book_data[ix]
            props = self.conn.msgproperties(payload=book, priority=priority)
            await queue.enqone(props)

        queue.deqoptions.condition = "priority = 9"
        messages = await queue.deqmany(3)
        self.assertEqual(len(messages), 0)

        for priority, ix in zip(priorities, indexes):
            queue.deqoptions.condition = f"priority = {priority}"
            messages = await queue.deqmany(3)
            self.assertEqual(len(messages), 1)
            book = messages[0].payload
            data = book.TITLE, book.AUTHORS, book.PRICE
            self.assertEqual(data, self.book_data[ix])

    async def test_8423(self):
        "8423 - test mode deqoption with DEQ_REMOVE_NODATA"
        queue = await self.get_and_clear_queue(
            self.book_queue_name, self.book_type_name
        )
        queue.deqoptions.wait = oracledb.DEQ_NO_WAIT
        queue.deqoptions.mode = oracledb.DEQ_REMOVE_NODATA

        book = queue.payload_type.newobject()
        for data in self.book_data:
            book.TITLE, book.AUTHORS, book.PRICE = data
            props = self.conn.msgproperties(payload=book)
            await queue.enqone(props)

        messages = await queue.deqmany(5)
        self.assertEqual(len(messages), 3)
        for message in messages:
            self.assertIsNone(message.payload.TITLE)
            self.assertIsNone(message.payload.AUTHORS)
            self.assertIsNone(message.payload.PRICE)

    async def test_8424(self):
        "8424 - test payload_type returns the correct value"
        books_type = await self.conn.gettype(self.book_type_name)
        queue = self.conn.queue(self.book_queue_name, books_type)
        self.assertEqual(queue.payload_type, books_type)

    async def test_8425(self):
        "8425 - test enqueuing to an object queue with the wrong payload"
        queue = await self.get_and_clear_queue(
            self.book_queue_name, self.book_type_name
        )
        props = self.conn.msgproperties(payload="A string")
        with self.assertRaisesFullCode("DPY-2062"):
            await queue.enqone(props)


if __name__ == "__main__":
    test_env.run_test_cases()



#------------------------------------------------------------------------------
# Copyright (c) 2023, 2024, Oracle and/or its affiliates.
#
# This software is dual-licensed to you under the Universal Permissive License
# (UPL) 1.0 as shown at https://oss.oracle.com/licenses/upl and Apache License
# 2.0 as shown at http://www.apache.org/licenses/LICENSE-2.0. You may choose
# either license.
#
# If you elect to accept the software under the Apache License, Version 2.0,
# the following applies:
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#------------------------------------------------------------------------------

#------------------------------------------------------------------------------
# vector.pyx
#
# Cython file defining the classes and methods used for encoding and decoding
# VECTOR data (embedded in base_impl.pyx).
#------------------------------------------------------------------------------

cdef array.array float_template = array.array('f')
cdef array.array double_template = array.array('d')
cdef array.array int8_template = array.array('b')
cdef array.array uint8_template = array.array('B')

@cython.final
cdef class SparseVectorImpl:

    @classmethod
    def from_values(cls, num_dimensions, indices, values):
        """
        Creates an implementation from its component values.
        """
        cdef SparseVectorImpl impl = cls.__new__(cls)
        impl.num_dimensions = num_dimensions
        impl.indices = indices
        impl.values = values
        return impl


@cython.final
cdef class VectorDecoder(Buffer):

    cdef array.array _decode_values(self, uint32_t num_elements,
                                    uint8_t vector_format):
        """
        Returns an array containing the decoded values.
        """
        cdef:
            uint8_t *uint8_buf = NULL
            double *double_buf = NULL
            uint8_t element_size = 0
            int8_t *int8_buf = NULL
            float *float_buf = NULL
            OracleDataBuffer buffer
            array.array result
            uint32_t i

        # set up buffers based on vector storage format
        if vector_format == VECTOR_FORMAT_FLOAT32:
            result = array.clone(float_template, num_elements, False)
            float_buf = result.data.as_floats
        elif vector_format == VECTOR_FORMAT_FLOAT64:
            result = array.clone(double_template, num_elements, False)
            double_buf = result.data.as_doubles
        elif vector_format == VECTOR_FORMAT_INT8:
            result = array.clone(int8_template, num_elements, False)
            int8_buf = result.data.as_schars
        elif vector_format == VECTOR_FORMAT_BINARY:
            num_elements = num_elements // 8
            result = array.clone(uint8_template, num_elements, False)
            uint8_buf = result.data.as_uchars
        else:
            errors._raise_err(errors.ERR_VECTOR_FORMAT_NOT_SUPPORTED,
                              vector_format=vector_format)

        # parse data
        for i in range(num_elements):
            if vector_format == VECTOR_FORMAT_FLOAT32:
                decode_binary_float(self._get_raw(4), 4, &buffer)
                float_buf[i] = buffer.as_float
            elif vector_format == VECTOR_FORMAT_FLOAT64:
                decode_binary_double(self._get_raw(8), 8, &buffer)
                double_buf[i] = buffer.as_double
            elif vector_format == VECTOR_FORMAT_INT8:
                self.read_sb1(&int8_buf[i])
            else:
                self.read_ub1(&uint8_buf[i])

        return result

    cdef object decode(self, bytes data):
        """
        Returns a Python object corresponding to the encoded VECTOR bytes.
        """
        cdef:
            uint8_t magic_byte, version, vector_format
            uint16_t flags, num_sparse_elements
            SparseVectorImpl sparse_impl
            array.array uint32_template
            uint32_t* sparse_indices
            uint32_t num_elements, i

        # populate the buffer with the data
        self._populate_from_bytes(data)

        # parse header
        self.read_ub1(&magic_byte)
        if magic_byte != TNS_VECTOR_MAGIC_BYTE:
            errors._raise_err(errors.ERR_UNEXPECTED_DATA,
                              data=bytes([magic_byte]))
        self.read_ub1(&version)
        if version > TNS_VECTOR_VERSION_WITH_SPARSE:
            errors._raise_err(errors.ERR_VECTOR_VERSION_NOT_SUPPORTED,
                              version=version)
        self.read_uint16be(&flags)
        self.read_ub1(&vector_format)
        self.read_uint32be(&num_elements)
        if flags & TNS_VECTOR_FLAG_NORM_RESERVED \
                or flags & TNS_VECTOR_FLAG_NORM:
            self.skip_raw_bytes(8)

        # for sparse vectors, only non-zero elements are found in the image
        if flags & TNS_VECTOR_FLAG_SPARSE:
            sparse_impl = SparseVectorImpl.__new__(SparseVectorImpl)
            sparse_impl.num_dimensions = num_elements
            self.read_uint16be(&num_sparse_elements)
            num_elements = num_sparse_elements
            uint32_template = array.array(ARRAY_TYPE_CODE_UINT32)
            sparse_impl.indices = array.clone(uint32_template,
                                              num_sparse_elements, False)
            sparse_indices = <uint32_t*> sparse_impl.indices.data.as_voidptr
            for i in range(num_sparse_elements):
                self.read_uint32be(&sparse_indices[i])
            sparse_impl.values = self._decode_values(num_sparse_elements,
                                                     vector_format)
            return PY_TYPE_SPARSE_VECTOR._from_impl(sparse_impl)

        # all other vectors have just the values
        return self._decode_values(num_elements, vector_format)


@cython.final
cdef class VectorEncoder(GrowableBuffer):

    cdef int _encode_values(self, array.array value, uint32_t num_elements,
                            uint8_t vector_format) except -1:
        """
        Encode the values into the image using the given vector storage format.
        """
        cdef:
            double *double_ptr = value.data.as_doubles
            uint8_t *uint8_ptr = value.data.as_uchars
            float *float_ptr = value.data.as_floats
            int8_t *int8_ptr = value.data.as_schars
            uint32_t i
        if vector_format == VECTOR_FORMAT_INT8:
            self.write_raw(<char_type*> int8_ptr, num_elements)
        elif vector_format == VECTOR_FORMAT_BINARY:
            self.write_raw(<char_type*> uint8_ptr, num_elements // 8)
        else:
            for i in range(num_elements):
                if vector_format == VECTOR_FORMAT_FLOAT32:
                    self.write_binary_float(float_ptr[i], write_length=False)
                elif vector_format == VECTOR_FORMAT_FLOAT64:
                    self.write_binary_double(double_ptr[i], write_length=False)

    cdef uint8_t _get_vector_format(self, array.array value):
        """
        Returns the vector storage format used by the array.
        """
        if value.typecode == 'd':
            return VECTOR_FORMAT_FLOAT64
        elif value.typecode == 'f':
            return VECTOR_FORMAT_FLOAT32
        elif value.typecode == 'b':
            return VECTOR_FORMAT_INT8
        return VECTOR_FORMAT_BINARY

    cdef int encode(self, object value) except -1:
        """
        Encodes the given value to the internal VECTOR format.
        """
        cdef:
            uint16_t flags = TNS_VECTOR_FLAG_NORM_RESERVED
            uint8_t vector_format, vector_version
            SparseVectorImpl sparse_impl = None
            uint16_t num_sparse_elements, i
            uint32_t* sparse_indices
            uint32_t num_elements

        # determine metadatda about the vector to write
        if isinstance(value, PY_TYPE_SPARSE_VECTOR):
            sparse_impl = value._impl
            num_elements = sparse_impl.num_dimensions
            vector_format = self._get_vector_format(sparse_impl.values)
            vector_version = TNS_VECTOR_VERSION_WITH_SPARSE
            flags |= TNS_VECTOR_FLAG_SPARSE | TNS_VECTOR_FLAG_NORM
        else:
            vector_format = self._get_vector_format(value)
            if vector_format == VECTOR_FORMAT_BINARY:
                num_elements = (<uint32_t> len(value)) * 8
                vector_version = TNS_VECTOR_VERSION_WITH_BINARY
            else:
                num_elements = <uint32_t> len(value)
                vector_version = TNS_VECTOR_VERSION_BASE
                flags |= TNS_VECTOR_FLAG_NORM

        # write header
        self.write_uint8(TNS_VECTOR_MAGIC_BYTE)
        self.write_uint8(vector_version)
        self.write_uint16be(flags)
        self.write_uint8(vector_format)
        self.write_uint32be(num_elements)
        self._reserve_space(8)              # reserve space for norm

        # write data
        if sparse_impl is None:
            self._encode_values(value, num_elements, vector_format)
        else:
            sparse_indices = <uint32_t*> sparse_impl.indices.data.as_voidptr
            num_sparse_elements = len(sparse_impl.indices)
            self.write_uint16be(num_sparse_elements)
            for i in range(num_sparse_elements):
                self.write_uint32be(sparse_indices[i])
            self._encode_values(sparse_impl.values, num_sparse_elements,
                                vector_format)



#------------------------------------------------------------------------------
# Copyright (c) 2023, 2024, Oracle and/or its affiliates.
#
# This software is dual-licensed to you under the Universal Permissive License
# (UPL) 1.0 as shown at https://oss.oracle.com/licenses/upl and Apache License
# 2.0 as shown at http://www.apache.org/licenses/LICENSE-2.0. You may choose
# either license.
#
# If you elect to accept the software under the Apache License, Version 2.0,
# the following applies:
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#------------------------------------------------------------------------------

#------------------------------------------------------------------------------
# transport.pyx
#
# Cython file defining the transport class used by the client for sending and
# receiving packets (embedded in thin_impl.pyx).
#------------------------------------------------------------------------------

cdef bint DEBUG_PACKETS = ("PYO_DEBUG_PACKETS" in os.environ)

cdef class Transport:

    cdef:
        object _transport
        object _ssl_context
        str _ssl_sni_data
        uint32_t _transport_num
        ssize_t _max_packet_size
        uint32_t _op_num
        bytes _partial_buf
        bint _full_packet_size
        bint _is_async

    cdef str _calc_sni_data(self, Description description):
        """
        Calculates the string used for the special SNI handling that allows one
        of the TLS negotiations to be bypassed.
        """
        cdef str server_type_part = ""
        if description.server_type is not None:
            server_type_part = f".T1.{description.server_type[:1]}"
        return (
            f"S{len(description.service_name)}.{description.service_name}"
            f"{server_type_part}"
            f".V3.{TNS_VERSION_DESIRED}"
        )

    cdef str _get_debugging_header(self, str operation):
        """
        Returns the header line used for debugging packets.
        """
        cdef str header
        self._op_num += 1
        current_date = datetime.datetime.now().isoformat(
            sep=" ",
            timespec="milliseconds"
        )
        return (
            f"{current_date} "
            f"{operation} [op {self._op_num}] on socket {self._transport_num}"
        )

    cdef int _print_output(self, str text) except -1:
        """
        Prints and flushes the text to stdout to ensure that multiple
        threads don't lose output due to buffering.
        """
        print(text + "\n", flush=True)

    cdef int _print_packet(self, str operation, object data) except -1:
        """
        Print the packet content in a format suitable for debugging.
        """
        offset = 0
        hex_data = memoryview(data).hex().upper()
        output_lines = [self._get_debugging_header(operation)]
        while hex_data:
            line_hex_data = hex_data[:16]
            hex_data = hex_data[16:]
            hex_dump_values = []
            printable_values = []
            for i in range(0, len(line_hex_data), 2):
                hex_byte = line_hex_data[i:i + 2]
                hex_dump_values.append(hex_byte)
                byte_val = ord(bytes.fromhex(hex_byte))
                char_byte = chr(byte_val)
                if char_byte.isprintable() and byte_val < 128 \
                        and char_byte != " ":
                    printable_values.append(char_byte)
                else:
                    printable_values.append(".")
            while len(hex_dump_values) < 8:
                hex_dump_values.append("  ")
                printable_values.append(" ")
            output_lines.append(
                f'{offset:04} : {" ".join(hex_dump_values)} '
                f'|{"".join(printable_values)}|'
            )
            offset += 8
        self._print_output("\n".join(output_lines))

    cdef int disconnect(self) except -1:
        """
        Disconnects the transport.
        """
        if self._transport is not None:
            if DEBUG_PACKETS:
                self._print_output(
                    self._get_debugging_header("Disconnecting transport")
                )
            self._transport.close()
            self._transport = None

    cdef int create_ssl_context(self, ConnectParamsImpl params,
                                Description description,
                                Address address) except -1:
        """
        Creates the SSL context used for establishing TLS communications
        between the database and the client.
        """

        # start with a default SSL context, unless a custom one is supplied
        self._ssl_context = params.ssl_context
        if self._ssl_context is None:
            self._ssl_context = ssl.create_default_context()

        # set the minimum and maximum versions of the TLS protocol
        if description.ssl_version is not None:
            self._ssl_context.minimum_version = description.ssl_version
            self._ssl_context.maximum_version = description.ssl_version

        # if the platform is macOS, check if the certifi package is installed;
        # if certifi is not installed, load the certificates from the macOS
        # keychain in PEM format
        if sys.platform == "darwin" and certifi is None:
            global macos_certs
            if macos_certs is None:
                certs = subprocess.run(["security", "find-certificate",
                                        "-a", "-p"],
                                        stdout=subprocess.PIPE).stdout
                macos_certs = certs.decode("utf-8")
            self._ssl_context.load_verify_locations(cadata=macos_certs)

        # if a wallet is specified, either mTLS is being used or a set of
        # certificates is being loaded to validate the server
        if description.wallet_location is not None:
            pem_file_name = os.path.join(description.wallet_location,
                                         PEM_WALLET_FILE_NAME)
            if not os.path.exists(pem_file_name):
                errors._raise_err(errors.ERR_WALLET_FILE_MISSING,
                                  name=pem_file_name)
            self._ssl_context.load_verify_locations(pem_file_name)
            wallet_password = params._get_wallet_password()
            try:
                self._ssl_context.load_cert_chain(pem_file_name,
                                                  password=wallet_password)
            except ssl.SSLError:
                pass

        # the SSL host name is not checked but the SSL server DN matching
        # algorithm is run instead after the TLS connection has been
        # established
        self._ssl_context.check_hostname = False

        # calculate the SNI data to send to the server, if applicable
        if description.use_sni:
            self._ssl_sni_data = self._calc_sni_data(description)
        else:
            self._ssl_sni_data = None

    cdef Packet extract_packet(self, bytes data=None):
        """
        Extracts a packet from the data, if possible (after first appending it
        to the partial buffer, if applicable). Any extra data not needed by the
        packet is stored in the partial buffer for a later call to this
        function.
        """
        cdef:
            ssize_t size, packet_size
            Packet packet
            char_type *ptr

        # update the partial buffer with the new data
        if data is not None:
            if self._partial_buf is None:
                self._partial_buf = data
            else:
                self._partial_buf += data
        size = 0 if self._partial_buf is None else len(self._partial_buf)

        # if enough bytes for the packet header, extract the packet size
        if size >= PACKET_HEADER_SIZE:

            # extract the packet size
            ptr = <char_type*> self._partial_buf
            if self._full_packet_size:
                packet_size = decode_uint32be(ptr)
            else:
                packet_size = decode_uint16be(ptr)

            # if enough bytes are available for the packet, return it
            if size >= packet_size:
                packet = Packet.__new__(Packet)
                packet.packet_size = packet_size
                packet.packet_type = ptr[4]
                packet.packet_flags = ptr[5]

                # store the packet buffer and adjust the partial buffer, as
                # needed
                if size == packet_size:
                    packet.buf = self._partial_buf
                    self._partial_buf = None
                else:
                    packet.buf = self._partial_buf[:packet_size]
                    self._partial_buf = self._partial_buf[packet_size:]

                # display packet, if requested
                if DEBUG_PACKETS:
                    self._print_packet("Receiving packet", packet.buf)
                return packet

    cdef tuple get_host_info(self):
        """
        Return a 2-tuple supplying the host and port to which the transport is
        connected.
        """
        cdef object sock
        if self._is_async:
            sock = self._transport.get_extra_info('socket')
        else:
            sock = self._transport
        return sock.getpeername()[:2]

    cdef int has_data_ready(self, bint *data_ready) except -1:
        """
        Returns true if data is ready to be read on the transport.
        """
        socket_list = [self._transport]
        read_socks, _, _ = select.select(socket_list, [], [], 0)
        data_ready[0] = bool(read_socks)

    cdef int negotiate_tls(self, object sock, Address address,
                           Description description) except -1:
        """
        Negotiate TLS on the socket.
        """
        if DEBUG_PACKETS:
            self._print_output(self._get_debugging_header("Negotiate TLS"))
        self._transport = self._ssl_context.wrap_socket(
            sock, server_hostname=self._ssl_sni_data
        )
        if description.ssl_server_dn_match:
            check_server_dn(self._transport, description.ssl_server_cert_dn,
                            address.host)

    cdef int renegotiate_tls(self, Address address,
                             Description description) except -1:
        """
        Renegotiate TLS on the socket.
        """
        orig_sock = self._transport
        sock = socket.socket(family=orig_sock.family, type=orig_sock.type,
                             proto=orig_sock.proto, fileno=orig_sock.detach())
        self.negotiate_tls(sock, address, description)

    async def negotiate_tls_async(self, BaseAsyncProtocol protocol,
                                  Address address, Description description):
        """
        Negotiate TLS on the socket asynchronously.
        """
        if DEBUG_PACKETS:
            self._print_output(self._get_debugging_header("Negotiate TLS"))
        orig_transport = self._transport
        loop = protocol._read_buf._loop
        self._transport = await loop.start_tls(
            self._transport, protocol,
            self._ssl_context,
            server_hostname=self._ssl_sni_data
        )
        if description.ssl_server_dn_match:
            sock = self._transport.get_extra_info("ssl_object")
            check_server_dn(sock, description.ssl_server_cert_dn, address.host)
        return orig_transport

    cdef int send_oob_break(self) except -1:
        """
        Sends an out-of-band break on the transport.
        """
        if DEBUG_PACKETS:
            self._print_output(
                self._get_debugging_header("Sending out of band break")
            )
        self._transport.send(b"!", socket.MSG_OOB)

    cdef int set_from_socket(self, object transport,
                             ConnectParamsImpl params,
                             Description description,
                             Address address) except -1:
        """
        Sets the transport from a socket.
        """
        cdef object sock
        if self._is_async:
            sock = transport.get_extra_info('socket')
        else:
            sock = transport
        if description.expire_time > 0:
            sock.setsockopt(socket.SOL_SOCKET, socket.SO_KEEPALIVE, 1)
            if hasattr(socket, "TCP_KEEPIDLE") \
                    and hasattr(socket, "TCP_KEEPINTVL") \
                    and hasattr(socket, "TCP_KEEPCNT"):
                sock.setsockopt(socket.IPPROTO_TCP, socket.TCP_KEEPIDLE,
                                description.expire_time * 60)
                sock.setsockopt(socket.IPPROTO_TCP, socket.TCP_KEEPINTVL, 6)
                sock.setsockopt(socket.IPPROTO_TCP, socket.TCP_KEEPCNT, 10)
        sock.setsockopt(socket.IPPROTO_TCP, socket.TCP_NODELAY, 1)
        if sock.gettimeout() != 0:
            sock.settimeout(None)
        self._transport = transport
        self._transport_num = sock.fileno()

    cdef Packet read_packet(self):
        """
        Reads a packet from the transport.
        """
        cdef:
            Packet packet
            bytes data
        packet = self.extract_packet()
        while packet is None:
            try:
                data = self._transport.recv(self._max_packet_size)
            except ConnectionResetError as e:
                errors._raise_err(errors.ERR_CONNECTION_CLOSED, str(e),
                                  cause=e)
            if len(data) == 0:
                self.disconnect()
                errors._raise_err(errors.ERR_CONNECTION_CLOSED)
            packet = self.extract_packet(data)
        return packet

    cdef int set_timeout(self, double value) except -1:
        """
        Sets the timeout on the transport.
        """
        self._transport.settimeout(value or None)

    cdef int write_packet(self, WriteBuffer buf) except -1:
        """
        Writes a packet on the transport.
        """
        data = bytes(buf._data_view[:buf._pos])
        if DEBUG_PACKETS:
            self._print_packet("Sending packet", data)
        try:
            if self._is_async:
                self._transport.write(data)
            else:
                self._transport.send(data)
        except OSError as e:
            self.disconnect()
            errors._raise_err(errors.ERR_CONNECTION_CLOSED, cause=e)



# -----------------------------------------------------------------------------
# Copyright (c) 2024, 2025, Oracle and/or its affiliates.
#
# This software is dual-licensed to you under the Universal Permissive License
# (UPL) 1.0 as shown at https://oss.oracle.com/licenses/upl and Apache License
# 2.0 as shown at http://www.apache.org/licenses/LICENSE-2.0. You may choose
# either license.
#
# If you elect to accept the software under the Apache License, Version 2.0,
# the following applies:
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# -----------------------------------------------------------------------------

# -----------------------------------------------------------------------------
# oci_tokens.py
#
# Methods that generates an OCI access token using the OCI SDK
# -----------------------------------------------------------------------------

import oci
import oracledb
from cryptography.hazmat.primitives.asymmetric import rsa
from cryptography.hazmat.primitives import serialization


def generate_token(token_auth_config, refresh=False):
    """
    Generates an OCI access token based on provided credentials.
    """
    user_auth_type = token_auth_config.get("auth_type") or ""
    auth_type = user_auth_type.lower()
    if auth_type == "configfileauthentication":
        return _config_file_based_authentication(token_auth_config)
    elif auth_type == "simpleauthentication":
        return _simple_authentication(token_auth_config)
    else:
        raise ValueError(
            f"Unrecognized auth_type authentication method {user_auth_type}"
        )


def _get_key_pair():
    """
    Generates a public-private key pair for proof of possession.
    """
    private_key = rsa.generate_private_key(
        public_exponent=65537,
        key_size=4096,
    )
    private_key_pem = private_key.private_bytes(
        encoding=serialization.Encoding.PEM,
        format=serialization.PrivateFormat.PKCS8,
        encryption_algorithm=serialization.NoEncryption(),
    ).decode("utf-8")

    public_key_pem = (
        private_key.public_key()
        .public_bytes(
            encoding=serialization.Encoding.PEM,
            format=serialization.PublicFormat.SubjectPublicKeyInfo,
        )
        .decode("utf-8")
    )

    if not oracledb.is_thin_mode():
        p_key = "".join(
            line.strip()
            for line in private_key_pem.splitlines()
            if not (
                line.startswith("-----BEGIN") or line.startswith("-----END")
            )
        )
        private_key_pem = p_key

    return {"private_key": private_key_pem, "public_key": public_key_pem}


def _config_file_based_authentication(token_auth_config):
    """
    Config file base authentication implementation: config parameters
    are provided in a file.
    """
    file_location = token_auth_config.get(
        "file_location", oci.config.DEFAULT_LOCATION
    )
    profile = token_auth_config.get("profile", oci.config.DEFAULT_PROFILE)

    # Load OCI config
    config = oci.config.from_file(file_location, profile)
    oci.config.validate_config(config)

    # Initialize service client with default config file
    client = oci.identity_data_plane.DataplaneClient(config)

    key_pair = _get_key_pair()

    response = client.generate_scoped_access_token(
        generate_scoped_access_token_details=oci.identity_data_plane.models.GenerateScopedAccessTokenDetails(
            scope="urn:oracle:db::id::*", public_key=key_pair["public_key"]
        )
    )

    # access_token is a tuple holding token and private key
    access_token = (
        response.data.token,
        key_pair["private_key"],
    )

    return access_token


def _simple_authentication(token_auth_config):
    """
    Simple authentication: config parameters are passed as parameters
    """
    config = {
        "user": token_auth_config["user"],
        "key_file": token_auth_config["key_file"],
        "fingerprint": token_auth_config["fingerprint"],
        "tenancy": token_auth_config["tenancy"],
        "region": token_auth_config["region"],
        "profile": token_auth_config["profile"],
    }
    oci.config.validate_config(config)

    # Initialize service client with given configuration
    client = oci.identity_data_plane.DataplaneClient(config)

    key_pair = _get_key_pair()

    response = client.generate_scoped_access_token(
        generate_scoped_access_token_details=oci.identity_data_plane.models.GenerateScopedAccessTokenDetails(
            scope="urn:oracle:db::id::*", public_key=key_pair["public_key"]
        )
    )

    # access_token is a tuple holding token and private key
    access_token = (
        response.data.token,
        key_pair["private_key"],
    )

    return access_token


def oci_token_hook(params: oracledb.ConnectParams):
    """
    OCI-specific hook for generating a token.
    """
    if params.extra_auth_params is not None:

        def token_callback(refresh):
            return generate_token(params.extra_auth_params, refresh)

        params.set(access_token=token_callback)


# Register the token hook for OCI
oracledb.register_params_hook(oci_token_hook)



# -----------------------------------------------------------------------------
# Copyright (c) 2024, 2025, Oracle and/or its affiliates.
#
# This software is dual-licensed to you under the Universal Permissive License
# (UPL) 1.0 as shown at https://oss.oracle.com/licenses/upl and Apache License
# 2.0 as shown at http://www.apache.org/licenses/LICENSE-2.0. You may choose
# either license.
#
# If you elect to accept the software under the Apache License, Version 2.0,
# the following applies:
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# -----------------------------------------------------------------------------

# -----------------------------------------------------------------------------
# config_provider.py
#
# Contains the built-in config providers.
# -----------------------------------------------------------------------------

import base64
import json
import os
import urllib.parse
import warnings

from . import errors
from .utils import register_password_type, register_protocol


def config_provider_file_hook(protocol, protocol_arg, connect_params):
    """
    Hook for "config-file://". The protocol_arg is expected to be the name of a
    file containing one or more configurations. An optional "key" parameter is
    allowed which will choose a configuration from a set of configurations
    stored in the file.
    """
    pos = protocol_arg.find("?")
    if pos < 0:
        file_name = protocol_arg
        key = None
    else:
        file_name = protocol_arg[:pos]
        args = urllib.parse.parse_qs(protocol_arg[pos + 1 :])
        key = args.get("key")
        if key is not None:
            key = key[0]
    if not os.path.isabs(file_name):
        if connect_params.config_dir is None:
            errors._raise_err(errors.ERR_NO_CONFIG_DIR)
        file_name = os.path.join(connect_params.config_dir, file_name)
    config = json.load(open(file_name))
    if key is not None:
        config = config[key]
    connect_params.set_from_config(config)


register_protocol("config-file", config_provider_file_hook)


def ldap_hook(protocol, arg, params):
    """
    Default hook for LDAP which simply points the user to the documentation
    which explains how they can write their own hook for LDAP.
    This hook is needed for python-oracledb Thin mode,or when
    defaults.thick_mode_dsn_passthrough is False in Thick mode.
    """
    doc_url = (
        "https://python-oracledb.readthedocs.io/en/latest"
        "/user_guide/connection_handling.html#ldap-directory-naming"
    )
    message = (
        f"To use an LDAP URL in python-oracledb, "
        f"register an LDAP resolution function as shown in {doc_url}"
    )
    raise Exception(message)


register_protocol("ldap", ldap_hook)
register_protocol("ldaps", ldap_hook)


def password_type_base64_hook(args):
    """
    Hook for password type "base64". The key "value" in the supplied args is
    expected to be a base64-encoded string.
    """
    warnings.warn("base64 encoded passwords are insecure")
    return base64.b64decode(args["value"].encode()).decode()


register_password_type("base64", password_type_base64_hook)



#------------------------------------------------------------------------------
# Copyright (c) 2025, Oracle and/or its affiliates.
#
# This software is dual-licensed to you under the Universal Permissive License
# (UPL) 1.0 as shown at https://oss.oracle.com/licenses/upl and Apache License
# 2.0 as shown at http://www.apache.org/licenses/LICENSE-2.0. You may choose
# either license.
#
# If you elect to accept the software under the Apache License, Version 2.0,
# the following applies:
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#------------------------------------------------------------------------------

#------------------------------------------------------------------------------
# nanoarrow_bridge.pxd
#
# Cython definition file declaring the classes used for bridging between the
# nanoarrow C interface and Python.
#------------------------------------------------------------------------------

# cython: language_level = 3

from libc.stdint cimport int8_t, uint8_t, int16_t, uint16_t
from libc.stdint cimport int32_t, uint32_t, int64_t, uint64_t

cdef extern from "nanoarrow.h":

    cdef struct ArrowArray:
        int64_t length
        int64_t null_count
        int64_t offset
        int64_t n_buffers
        int64_t n_children
        ArrowArray** children
        const void** buffers
        void (*release)(ArrowArray*)

    cdef struct ArrowSchema:
        void (*release)(ArrowSchema*)

    cpdef enum ArrowType:
        NANOARROW_TYPE_BOOL
        NANOARROW_TYPE_BINARY
        NANOARROW_TYPE_DECIMAL128
        NANOARROW_TYPE_DOUBLE
        NANOARROW_TYPE_FLOAT
        NANOARROW_TYPE_INT64
        NANOARROW_TYPE_LARGE_BINARY
        NANOARROW_TYPE_LARGE_STRING
        NANOARROW_TYPE_STRING
        NANOARROW_TYPE_TIMESTAMP
        NANOARROW_TYPE_UNINITIALIZED

    cpdef enum ArrowTimeUnit:
        NANOARROW_TIME_UNIT_SECOND
        NANOARROW_TIME_UNIT_MILLI
        NANOARROW_TIME_UNIT_MICRO
        NANOARROW_TIME_UNIT_NANO


cdef class OracleArrowArray:
    """
    OracleArrowArray corresponds to a Column in the Relational model

    It uses functions defined in the Arrow C Data Interface
    to work with Arrow buffers and incrementally append values

    The only user-facing API in this object will be __arrow_c_array__()
    which is documented in the Arrow PyCapsule Interface. Arrow-backed
    DataFrame libraries will use __arrow_c_array__() to directly access
    the underlying arrow data

    """
    cdef:
        public int32_t precision
        public int32_t scale
        public str name
        public ArrowType arrow_type
        public ArrowTimeUnit time_unit
        double factor
        ArrowArray *arrow_array
        ArrowSchema *arrow_schema

    cdef str _schema_to_string(self)
    cdef int append_bytes(self, void* ptr, int64_t num_bytes) except -1
    cdef int append_decimal(self, void* ptr, int64_t num_bytes) except -1
    cdef int append_double(self, double value) except -1
    cdef int append_float(self, float value) except -1
    cdef int append_int64(self, int64_t value) except -1
    cdef int append_last_value(self, OracleArrowArray array) except -1
    cdef int append_null(self) except -1
    cdef int finish_building(self) except -1



# -----------------------------------------------------------------------------
# Copyright (c) 2024, 2025, Oracle and/or its affiliates.
#
# This software is dual-licensed to you under the Universal Permissive License
# (UPL) 1.0 as shown at https://oss.oracle.com/licenses/upl and Apache License
# 2.0 as shown at http://www.apache.org/licenses/LICENSE-2.0. You may choose
# either license.
#
# If you elect to accept the software under the Apache License, Version 2.0,
# the following applies:
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# -----------------------------------------------------------------------------

# -----------------------------------------------------------------------------
# azure_config_provider.py
#
# Python file contains the hook method config_azure_hook() that fetches config
# store from Azure App Configuration.
# -----------------------------------------------------------------------------

import json
import re

import oracledb

from urllib.parse import urlparse, parse_qs
from azure.appconfiguration import AzureAppConfigurationClient
from azure.keyvault.secrets import SecretClient
from azure.core.exceptions import ResourceNotFoundError
from azure.identity import (
    ClientSecretCredential,
    CertificateCredential,
    ManagedIdentityCredential,
    ChainedTokenCredential,
    EnvironmentCredential,
)


def _get_authentication_method(parameters):
    auth_method = parameters.get("authentication", parameters.get("method"))
    if auth_method is not None:
        auth_method = auth_method.upper()
        if auth_method == "AZURE_DEFAULT":
            auth_method = None
    return auth_method


def _get_credential(parameters):
    """
    Returns the appropriate credential given the input supplied by the original
    connect string.
    """

    tokens = []
    auth_method = _get_authentication_method(parameters)

    if auth_method is None or auth_method == "AZURE_SERVICE_PRINCIPAL":
        if "azure_client_secret" in parameters:
            tokens.append(
                ClientSecretCredential(
                    _get_required_parameter(parameters, "azure_tenant_id"),
                    _get_required_parameter(parameters, "azure_client_id"),
                    _get_required_parameter(parameters, "azure_client_secret"),
                )
            )
        elif "azure_client_certificate_path" in parameters:
            tokens.append(
                CertificateCredential(
                    _get_required_parameter(parameters, "azure_tenant_id"),
                    _get_required_parameter(parameters, "azure_client_id"),
                    _get_required_parameter(
                        parameters, "azure_client_certificate_path"
                    ),
                )
            )
    if auth_method is None or auth_method == "AZURE_MANAGED_IDENTITY":
        client_id = parameters.get("azure_managed_identity_client_id")
        if client_id is not None:
            tokens.append(ManagedIdentityCredential(client_id=client_id))

    if len(tokens) == 0:
        message = (
            "Authentication options were not available in Connection String"
        )
        raise Exception(message)
    elif len(tokens) == 1:
        return tokens[0]
    tokens.append(EnvironmentCredential())
    return ChainedTokenCredential(*tokens)


def _get_password(pwd_string, parameters):
    try:
        pwd = json.loads(pwd_string)
    except json.JSONDecodeError:
        message = (
            "Password is expected to be JSON"
            " containing Azure Vault details."
        )
        raise Exception(message)

    pwd["value"] = pwd.pop("uri")
    pwd["type"] = "azurevault"

    # make authentication section
    pwd["authentication"] = authentication = {}

    authentication["method"] = auth_method = _get_authentication_method(
        parameters
    )

    if auth_method is None or auth_method == "AZURE_SERVICE_PRINCIPAL":
        if "azure_client_secret" in parameters:
            authentication["azure_tenant_id"] = _get_required_parameter(
                parameters, "azure_tenant_id"
            )
            authentication["azure_client_id"] = _get_required_parameter(
                parameters, "azure_client_id"
            )
            authentication["azure_client_secret"] = _get_required_parameter(
                parameters, "azure_client_secret"
            )

        elif "azure_client_certificate_path" in parameters:
            authentication["azure_tenant_id"] = (
                _get_required_parameter(parameters, "azure_tenant_id"),
            )
            authentication["azure_client_id"] = (
                _get_required_parameter(parameters, "azure_client_id"),
            )
            authentication["azure_client_certificate_path"] = (
                _get_required_parameter(
                    parameters, "azure_client_certificate_path"
                )
            )

    if auth_method is None or auth_method == "AZURE_MANAGED_IDENTITY":
        authentication["azure_managed_identity_client_id"] = parameters.get(
            "azure_managed_identity_client_id"
        )
    return pwd


def _get_required_parameter(parameters, name, location="connection string"):
    try:
        return parameters[name]
    except KeyError:
        message = f'Parameter named "{name}" is missing from {location}'
        raise Exception(message) from None


def _get_setting(client, key, sub_key, label, required=True):
    """
    Returns the configuration setting given the client, key and label.
    """
    try:
        if key.endswith("/"):
            actual_key = f"{key}{sub_key}"
        else:
            actual_key = f"{key}/{sub_key}"
        obj = client.get_configuration_setting(key=actual_key, label=label)
    except ResourceNotFoundError:
        if required:
            message = f"Missing required configuration key: {actual_key}"
            raise Exception(message)
        return None
    return obj.value


def _parse_parameters(protocol_arg: str) -> dict:
    """
    Parse the parameters from the protocol argument string.
    """
    pos = protocol_arg.find("?")
    parsed_url = urlparse(protocol_arg[pos + 1 :])
    parsed_values = parse_qs(parsed_url.path)
    parameters = {
        key.lower(): value[0] for key, value in parsed_values.items()
    }
    config_name = protocol_arg[:pos].rstrip("/")
    if not config_name.endswith(".azconfig.io"):
        config_name += ".azconfig.io"
    parameters["appconfigname"] = config_name
    return parameters


def password_type_azure_vault_hook(args):
    uri = _get_required_parameter(args, "value", '"password" key section')
    credential = args.get("credential")

    if credential is None:
        # if credential not present, this might be coming
        # from oci config provider, so create credential
        # for azure key vault.
        auth = args.get("authentication")
        if auth is None:
            raise Exception(
                "Azure Vault authentication details were not provided."
            )
        credential = _get_credential(auth)

    pattern = re.compile(
        r"(?P<vault_url>https://[A-Za-z0-9._-]+)/"
        r"secrets/(?P<secretKey>[A-Za-z][A-Za-z0-9-]*)$"
    )
    match = pattern.match(uri)
    if match is None:
        raise Exception("Invalid Azure Vault details")
    vault_url = match.group("vault_url")
    secret_key = match.group("secretKey")
    secret_client = SecretClient(vault_url, credential)
    return secret_client.get_secret(secret_key).value


def _process_config(parameters, connect_params):
    """
    Processes the configuration stored in the Azure App configuration store.
    """

    credential = _get_credential(parameters)
    client = AzureAppConfigurationClient(
        "https://" + _get_required_parameter(parameters, "appconfigname"),
        credential,
    )
    key = _get_required_parameter(parameters, "key")
    label = parameters.get("label")

    # get the common parameters
    config = {}
    config["connect_descriptor"] = _get_setting(
        client, key, "connect_descriptor", label
    )
    config["user"] = _get_setting(client, key, "user", label, required=False)
    pwd = _get_setting(client, key, "password", label, required=False)
    if pwd is not None:
        config["password"] = _get_password(pwd, parameters)

    config["config_time_to_live"] = _get_setting(
        client, key, "config_time_to_live", label, required=False
    )
    config["config_time_to_live_grace_period"] = _get_setting(
        client, key, "config_time_to_live_grace_period", label, required=False
    )

    # get the python-oracledb specific parameters
    settings = _get_setting(client, key, "pyo", label, required=False)
    if settings is not None:
        config["pyo"] = json.loads(settings)

    # set the configuration
    connect_params.set_from_config(config)


def config_azure_hook(protocol, protocol_arg, connect_params):
    """
    Hook for handling parameters stored in an Azure configuration store.
    """
    parameters = _parse_parameters(protocol_arg)
    _process_config(parameters, connect_params)


oracledb.register_password_type("azurevault", password_type_azure_vault_hook)
oracledb.register_protocol("config-azure", config_azure_hook)
